import pytest
import responses

from cumulusci.tests.util import FakeUnreliableRequestHandler
from cumulusci.utils.http.multi_request import CompositeParallelSalesforce

COMPOSITE_RESPONSE = {
    "compositeResponse": [
        {
            "body": {
                "totalSize": 1,
                "done": True,
                "records": [
                    {
                        "attributes": {
                            "type": "Account",
                            "url": "/services/data/v49.0/sobjects/Account/0013B00000ddh3PQAQ",
                        },
                        "Id": "0013B00000ddh3PQAQ",
                    }
                ],
            },
            "httpHeaders": {},
            "httpStatusCode": 200,
            "referenceId": "one",
        },
        {
            "body": {
                "totalSize": 1,
                "done": True,
                "records": [
                    {
                        "attributes": {
                            "type": "Account",
                            "url": "/services/data/v49.0/sobjects/Account/0013B00000ddh7qQAA",
                        },
                        "Name": "Elizabeth Foster",
                    }
                ],
            },
            "httpHeaders": {},
            "httpStatusCode": 200,
            "referenceId": "two",
        },
    ]
}


class TestCompositeParallelSalesforce:
    @pytest.mark.vcr()
    def test_composite_parallel_salesforce(
        self, sf, run_code_without_recording, delete_data_from_org
    ):
        run_code_without_recording(lambda: delete_data_from_org("Entitlement,Account"))
        sf.Account.create(
            {"Name": "Smith Corp."},
        )
        requests = [
            {
                "method": "GET",
                "url": "/services/data/v50.0/query?q=SELECT Id FROM Account LIMIT 1",
            },
            {
                "method": "GET",
                "url": "/services/data/v50.0/query?q=SELECT Name FROM Account LIMIT 1",
            },
        ] * 2
        with CompositeParallelSalesforce(sf, 5, max_workers=1) as cpsf:
            results, errors = cpsf.do_composite_requests(requests)

        assert len(list(results)) == len(requests)
        for result in results:
            assert result["body"]
            assert result["body"]["done"]
            assert result["body"]["records"][0]["attributes"]

    @pytest.mark.vcr()
    def test_not_opened(self, sf):
        cpsf = CompositeParallelSalesforce(sf, 5, max_workers=1)
        with pytest.raises(AssertionError):
            cpsf.do_composite_requests([])

    @pytest.mark.vcr()
    def test_empty(self, sf):
        with CompositeParallelSalesforce(sf, 5, max_workers=1) as cpsf:
            results, errors = cpsf.do_composite_requests([])
            assert list(results) == []

    @pytest.mark.vcr()
    def test_http_headers(self, sf, vcr):
        requests = [
            {
                "method": "GET",
                "url": "/services/data/v50.0/sobjects",
                "httpHeaders": {"If-Modified-Since": "Thu, 03 Sep 2020 21:35:07 GMT"},
            },
        ] * 3
        # don't re-record this one because you'll need to fiddle
        # with the dates
        with vcr.use_cassette(
            "ManualEditTestCompositeParallelSalesforce.test_http_headers.yaml"
        ):
            with CompositeParallelSalesforce(sf, 4, max_workers=1) as cpsf:
                results, errors = cpsf.do_composite_requests(requests)
        assert results[0]["httpStatusCode"] == 304

    @pytest.mark.vcr()
    def test_reference_ids(self, sf, run_code_without_recording, delete_data_from_org):
        run_code_without_recording(lambda: delete_data_from_org("Account"))
        requests = [
            {
                "method": "GET",
                "url": "/services/data/v50.0/query?q=SELECT Id FROM Account LIMIT 1",
                "referenceId": "one",
            },
            {
                "method": "GET",
                "url": "/services/data/v50.0/query?q=SELECT Name FROM Account LIMIT 1",
                "referenceId": "two",
            },
            {
                "method": "GET",
                "url": "/services/data/v50.0/query?q=SELECT Name FROM Account LIMIT 1",
                "referenceId": "three",
            },
        ]
        with CompositeParallelSalesforce(sf, 2, max_workers=1) as cpsf:
            results, errors = cpsf.do_composite_requests(requests)
        assert len(list(results)) == len(requests)
        assert set(result["referenceId"] for result in results) == set(
            request["referenceId"] for request in requests
        )

    @pytest.mark.vcr()
    def test_errors(self, sf):
        requests = [{"method": "GET", "url": "/services/data/v50.0/sobjects/Foo"}]
        with CompositeParallelSalesforce(sf, 4, max_workers=1) as cpsf:
            results, errors = cpsf.do_composite_requests(requests)
        assert results[0]["httpStatusCode"] == 404, results[0]["httpStatusCode"]

    @responses.activate
    def test_multirequest_timeout(self, sf):
        requests = [
            {
                "method": "GET",
                "url": "/services/data/v50.0/query?q=SELECT Id FROM Account LIMIT 1",
                "referenceId": "one",
            },
            {
                "method": "GET",
                "url": "/services/data/v50.0/query?q=SELECT Name FROM Account LIMIT 1",
                "referenceId": "two",
            },
            {
                "method": "GET",
                "url": "/services/data/v50.0/query?q=SELECT Name FROM Account LIMIT 1",
                "referenceId": "three",
            },
        ] * 2

        composite_handler = FakeUnreliableRequestHandler(COMPOSITE_RESPONSE)
        responses.add_callback(
            responses.POST,
            f"{sf.base_url}composite",
            callback=composite_handler.request_callback,
            content_type="application/json",
        )

        single_request_handler = FakeUnreliableRequestHandler(COMPOSITE_RESPONSE)
        responses.add_callback(
            responses.GET,
            f"{sf.base_url}query?q=SELECT%20Id%20FROM%20Account%20LIMIT%201",
            callback=single_request_handler.request_callback,
            content_type="application/json",
        )

        with CompositeParallelSalesforce(sf, 2, max_workers=1) as cpsf:
            results, errors = cpsf.do_composite_requests(requests)
        assert len(errors) == 1  # one of the single requests fails after retry
        assert len(results) == 5  # 4 succeed first time, one fails on retry

        assert single_request_handler.counter == 2
        assert composite_handler.counter == 3

    @responses.activate
    def test_multirequest_unknown_exception(self, sf):
        requests = [
            {
                "method": "GET",
                "url": "/services/data/v50.0/query?q=SELECT Id FROM Account LIMIT 1",
                "referenceId": "one",
            },
            {
                "method": "GET",
                "url": "/services/data/v50.0/query?q=SELECT Name FROM Account LIMIT 1",
                "referenceId": "two",
            },
            {
                "method": "GET",
                "url": "/services/data/v50.0/query?q=SELECT Name FROM Account LIMIT 1",
                "referenceId": "three",
            },
        ] * 2

        with CompositeParallelSalesforce(sf, 2, max_workers=1) as cpsf:
            results, errors = cpsf.do_composite_requests(requests)
        assert len(errors) == 3  # all should be rejected by Responses.
        assert len(results) == 0

    @responses.activate
    def test_multirequest_timeout_POSTs(self, sf):
        requests = [
            {
                "method": "GET",
                "url": "/services/data/v50.0/query?q=SELECT Id FROM Account LIMIT 1",
                "referenceId": "one",
            },
            {
                "method": "GET",
                "url": "/services/data/v50.0/query?q=SELECT Name FROM Account LIMIT 1",
                "referenceId": "two",
            },
            {
                "method": "POST",
                "url": "/services/data/v50.0/query?q=SELECT Name FROM Account LIMIT 1",
                "referenceId": "three",
            },
        ] * 2

        composite_handler = FakeUnreliableRequestHandler(COMPOSITE_RESPONSE)
        responses.add_callback(
            responses.POST,
            f"{sf.base_url}composite",
            callback=composite_handler.request_callback,
            content_type="application/json",
        )

        single_request_handler = FakeUnreliableRequestHandler(COMPOSITE_RESPONSE)
        responses.add_callback(
            responses.GET,
            f"{sf.base_url}query?q=SELECT%20Id%20FROM%20Account%20LIMIT%201",
            callback=single_request_handler.request_callback,
            content_type="application/json",
        )

        with CompositeParallelSalesforce(sf, 2, max_workers=1) as cpsf:
            results, errors = cpsf.do_composite_requests(requests)

        # Three chunks of 2. SUCCEED, FAIL, SUCCEED
        assert composite_handler.counter == 3

        # The GET in the failing chunk should be retried and succeed
        # The POST should not.
        assert len(errors) == 1, str(errors)
        assert single_request_handler.counter == 1



import hashlib
import unittest
from unittest import mock

from github3.git import Tree
from github3.repos import Repository

from cumulusci.core.exceptions import GithubException
from cumulusci.tasks.github.util import CommitDir
from cumulusci.utils import temporary_dir


class TestCommitDir(unittest.TestCase):
    def test_call(self):
        with temporary_dir() as d:
            repo = mock.Mock(spec=Repository)
            repo.owner = "SalesforceFoundation"
            repo.name = "TestRepo"
            repo.tree = mock.Mock()
            repo.tree.return_value = Tree(
                {
                    "url": "string",
                    "sha": "tree-ish-hash",
                    "tree": [
                        {
                            "type": "tree",
                            "mode": "100644",
                            "path": "dir",
                            "sha": "bogus1",
                        },
                        {
                            "type": "blob",
                            "mode": "100644",
                            "path": "file_outside_dir",
                            "sha": "bogus2",
                        },
                        {
                            "type": "blob",
                            "mode": "100644",
                            "path": "dir/deleteme",
                            "sha": "deletedfilesha",
                        },
                        {
                            "type": "blob",
                            "mode": "100644",
                            "path": "dir/unchanged",
                            "sha": hashlib.sha1(b"blob 0\0").hexdigest(),
                        },
                        {
                            "type": "blob",
                            "mode": "100644",
                            "path": "dir/modified",
                            "sha": "bogus3",
                        },
                    ],
                },
                None,
            )

            commit = CommitDir(repo)
            with open("unchanged", "w") as f:
                f.write("")
            with open("modified", "w") as f:
                f.write("modified")
            with open("new", "w") as f:
                f.write("new")
            with open(".hidden", "w") as f:
                pass
            commit(d, "main", "dir", dry_run=True)
            assert commit.new_tree_list == [
                {
                    "sha": "bogus2",
                    "mode": "100644",
                    "path": "file_outside_dir",
                    "size": None,
                    "type": "blob",
                },
                {
                    "sha": hashlib.sha1(b"blob 0\0").hexdigest(),
                    "mode": "100644",
                    "path": "dir/unchanged",
                    "size": None,
                    "type": "blob",
                },
                {
                    "sha": None,
                    "mode": "100644",
                    "path": "dir/modified",
                    "size": None,
                    "type": "blob",
                },
                {"path": "dir/new", "mode": "100644", "type": "blob", "sha": None},
            ]
            commit(d, "main", "dir", commit_message="msg")
        repo.create_commit.assert_called_once()

    def test_call__no_changes(self):
        with temporary_dir() as d:
            repo = mock.Mock(spec=Repository)
            repo.tree = mock.Mock(
                return_value=Tree(
                    {"url": "string", "sha": "tree-ish-hash", "tree": []}, None
                )
            )
            commit = CommitDir(repo)
            commit(d, "main", commit_message="msg")
        repo.create_commit.assert_not_called()

    def test_validate_dirs(self):
        repo = mock.Mock(spec=Repository)
        commit = CommitDir(repo)
        with self.assertRaises(GithubException):
            commit._validate_dirs("bogus", None)
        _, repo_dir = commit._validate_dirs(".", None)
        self.assertEqual("", repo_dir)
        _, repo_dir = commit._validate_dirs(".", "./test/")
        self.assertEqual("test", repo_dir)

    def test_call__error_creating_tree(self):
        with temporary_dir() as d:
            repo = mock.Mock(spec=Repository)
            repo.create_tree.return_value = None
            repo.tree = mock.Mock(
                return_value=Tree(
                    {"url": "string", "sha": "tree-ish-hash", "tree": []}, None
                )
            )
            with open("new", "w") as f:
                f.write("new")
            commit = CommitDir(repo)
            with self.assertRaises(GithubException):
                commit(d, "main", commit_message="msg")

    def test_call__error_creating_commit(self):
        with temporary_dir() as d:
            repo = mock.Mock(spec=Repository)
            repo.create_commit.return_value = None
            repo.tree = mock.Mock(
                return_value=Tree(
                    {"url": "string", "sha": "tree-ish-hash", "tree": []}, None
                )
            )
            with open("new", "w") as f:
                f.write("new")
            commit = CommitDir(repo)
            with self.assertRaises(GithubException):
                commit(d, "main", commit_message="msg")

    def test_call__error_updating_head(self):
        with temporary_dir() as d:
            repo = mock.Mock(spec=Repository)
            head = mock.Mock()
            head.update.return_value = None
            repo.ref.return_value = head
            repo.tree = mock.Mock(
                return_value=Tree(
                    {"url": "string", "sha": "tree-ish-hash", "tree": []}, None
                )
            )
            with open("new", "w") as f:
                f.write("new")
            commit = CommitDir(repo)
            with self.assertRaises(GithubException):
                commit(d, "main", commit_message="msg")

    def test_create_blob__handles_decode_error(self):
        repo = mock.Mock(spec=Repository)
        commit = CommitDir(repo)
        commit.dry_run = False
        self.assertTrue(commit._create_blob(b"\x9c", "local_path"))

    def test_create_blob__error(self):
        repo = mock.Mock(spec=Repository)
        repo.create_blob.return_value = None
        commit = CommitDir(repo)
        commit.dry_run = False
        with self.assertRaises(GithubException):
            self.assertTrue(commit._create_blob(b"", "local_path"))



import code
import contextlib
import pdb
import runpy
import sys
import traceback

import click
import requests
from rich.console import Console

import cumulusci
from cumulusci.core.debug import set_debug_mode
from cumulusci.core.exceptions import CumulusCIUsageError
from cumulusci.utils import get_cci_upgrade_command
from cumulusci.utils.logging import tee_stdout_stderr

from .error import error
from .flow import flow
from .logger import get_tempfile_logger, init_logger
from .org import org
from .plan import plan
from .project import project
from .runtime import CliRuntime, pass_runtime
from .service import service
from .task import task
from .utils import check_latest_version, get_installed_version, get_latest_final_version

SUGGEST_ERROR_COMMAND = (
    """Run this command for more information about debugging errors: cci error --help"""
)

USAGE_ERRORS = (CumulusCIUsageError, click.UsageError)


#
# Root command
#
def main(args=None):
    """Main CumulusCI CLI entry point.

    This runs as the first step in processing any CLI command.

    This wraps the `click` library in order to do some initialization and centralized error handling.
    """
    with contextlib.ExitStack() as stack:
        args = args or sys.argv
        # Check for updates _unless_ we've been asked to output JSON,
        # or if we're going to check anyway as part of the `version` command.
        is_version_command = len(args) > 1 and args[1] == "version"
        if "--json" not in args and not is_version_command:
            check_latest_version()

        # Only create logfiles for commands that are not `cci error`
        is_error_command = len(args) > 2 and args[1] == "error"
        tempfile_path = None
        if not is_error_command:
            logger, tempfile_path = get_tempfile_logger()
            stack.enter_context(tee_stdout_stderr(args, logger, tempfile_path))

        debug = "--debug" in args
        if debug:
            args.remove("--debug")

        with set_debug_mode(debug):
            try:
                runtime = CliRuntime(load_keychain=False)
            except Exception as e:
                handle_exception(e, is_error_command, tempfile_path, debug)
                sys.exit(1)

            runtime.check_cumulusci_version()
            should_show_stacktraces = runtime.universal_config.cli__show_stacktraces

            init_logger(debug=debug)
            # Hand CLI processing over to click, but handle exceptions
            try:
                cli(args[1:], standalone_mode=False, obj=runtime)
            except click.Abort:  # Keyboard interrupt
                console = Console()
                show_debug_info() if debug else console.print("\n[red bold]Aborted!")
                sys.exit(1)
            except Exception as e:
                if debug:
                    show_debug_info()
                else:
                    handle_exception(
                        e,
                        is_error_command,
                        tempfile_path,
                        should_show_stacktraces,
                    )
                sys.exit(1)


def handle_exception(
    error,
    is_error_cmd,
    logfile_path,
    should_show_stacktraces=False,
):
    """Displays error of appropriate message back to user, prompts user to investigate further
    with `cci error` commands, and writes the traceback to the latest logfile.
    """
    error_console = Console(stderr=True)
    if isinstance(error, requests.exceptions.ConnectionError):
        connection_error_message(error_console)
    elif isinstance(error, click.ClickException):
        error_console.print(f"[red bold]Error: {error.format_message()}")
    else:
        error_console.print(f"[red bold]Error: {error}")
    # Only suggest gist command if it wasn't run
    if not is_error_cmd:
        error_console.print(f"[yellow]{SUGGEST_ERROR_COMMAND}")

    # This is None if we're handling an exception for a `cci error` command.
    if logfile_path:
        with open(logfile_path, "a") as log_file:
            traceback.print_exc(file=log_file)  # log stacktrace silently

    if should_show_stacktraces and not isinstance(error, USAGE_ERRORS):
        raise error


def connection_error_message(console: Console):
    message = (
        "We encountered an error with your internet connection. "
        "Please check your connection and try the last cci command again."
    )
    console.print(f"[red bold]{message}")


def show_debug_info():
    """Displays the traceback and opens pdb"""
    traceback.print_exc()
    pdb.post_mortem()


@click.group("main", help="")
def cli():
    """Top-level `click` command group."""


@cli.command(name="version", help="Print the current version of CumulusCI")
def version():
    click.echo("CumulusCI version: ", nl=False)
    click.echo(click.style(cumulusci.__version__, bold=True), nl=False)
    click.echo(f" ({sys.argv[0]})")
    click.echo(f"Python version: {sys.version.split()[0]}", nl=False)
    click.echo(f" ({sys.executable})")

    click.echo()
    current_version = get_installed_version()
    latest_version = get_latest_final_version()
    if latest_version > current_version:
        click.echo(
            f"There is a newer version of CumulusCI available ({str(latest_version)})."
        )
        click.echo(f"To upgrade, run `{get_cci_upgrade_command()}`")
        click.echo(
            f"Release notes: https://github.com/SFDO-Tooling/CumulusCI/releases/tag/v{str(latest_version)}"
        )
    else:
        click.echo("You have the latest version of CumulusCI.")

    click.echo()


@cli.command(name="shell", help="Drop into a Python shell")
@click.option("--script", help="Path to a script to run", type=click.Path())
@click.option("--python", help="Python code to run directly")
@pass_runtime(require_project=False, require_keychain=True)
def shell(runtime, script=None, python=None):
    # alias for backwards-compatibility
    variables = {
        "config": runtime,
        "runtime": runtime,
        "project_config": runtime.project_config,
    }

    if script:
        if python:
            raise click.UsageError("Cannot specify both --script and --python")
        runpy.run_path(script, init_globals=variables)
    elif python:
        exec(python, variables)
    else:
        code.interact(local=variables)


# Top Level Groups

cli.add_command(error)
cli.add_command(project)
cli.add_command(org)
cli.add_command(service)
cli.add_command(task)
cli.add_command(flow)
cli.add_command(plan)



from unittest import mock

import pytest
from github3.exceptions import NotFoundError

from cumulusci.core.config import BaseConfig
from cumulusci.core.dependencies.github import (
    get_package_data,
    get_remote_project_config,
    get_repo,
)
from cumulusci.core.exceptions import DependencyResolutionError


class DummyResponse(object):
    status_code = 404
    content = ""


def test_get_remote_project_config():
    content = b"""
project:
    package:
        namespace: foo"""

    repo = mock.Mock()
    repo.file_contents.return_value.decoded = content
    assert isinstance(get_remote_project_config(repo, "aaaaaaaa"), BaseConfig)


def test_get_repo():
    context = mock.Mock()

    assert get_repo("test", context) == context.get_repo_from_url.return_value


def test_get_repo__failure():
    context = mock.Mock()
    context.get_repo_from_url.return_value = None

    with pytest.raises(DependencyResolutionError):
        get_repo("test", context)


def test_get_repo__404():
    context = mock.Mock()
    context.get_repo_from_url.side_effect = NotFoundError(DummyResponse)

    with pytest.raises(DependencyResolutionError):
        get_repo("test", context)


def test_get_package_data():
    content = b"""
project:
    package:
        namespace: foo"""

    repo = mock.Mock()
    repo.file_contents.return_value.decoded = content

    assert get_package_data(get_remote_project_config(repo, "aaaaaaaa")) == (
        "Package",
        "foo",
    )



from unittest import mock

import click
import pytest

from cumulusci.core.exceptions import ServiceNotConfigured
from cumulusci.core.tests.utils import EnvironmentVarGuard

from .. import service
from .utils import run_click_command


@mock.patch("cumulusci.cli.service.CliTable")
def test_service_list__no_active_defaults(cli_tbl):
    runtime = mock.Mock()
    runtime.project_config.services = {
        "bad": {"description": "Unconfigured Service"},
        "test": {"description": "Test Service"},
        "something_else": {"description": "something else"},
    }
    runtime.keychain.list_services.return_value = {
        "test": ["test_alias", "test2_alias"],
        "bad": ["bad_alias"],
    }
    runtime.keychain._default_services = {"test": "test_alias"}
    runtime.universal_config.cli__plain_output = None

    run_click_command(
        service.service_list, runtime=runtime, plain=False, print_json=False
    )

    cli_tbl.assert_called_with(
        [
            ["Default", "Type", "Name", "Description"],
            [False, "bad", "bad_alias", "Unconfigured Service"],
            [False, "something_else", "", "something else"],
            [True, "test", "test_alias", "Test Service"],
            [False, "test", "test2_alias", "Test Service"],
        ],
        title="Services",
        dim_rows=[2],
    )


@mock.patch("cumulusci.cli.service.CliTable")
def test_service_list(cli_tbl):
    runtime = mock.Mock()
    runtime.project_config.services = {
        "bad": {"description": "Unconfigured Service"},
        "test": {"description": "Test Service"},
        "something_else": {"description": "something else"},
    }
    runtime.keychain.list_services.return_value = {
        "test": ["test_alias", "test2_alias"],
        "bad": ["bad_alias"],
    }
    runtime.keychain._default_services = {"test": "test_alias", "bad": "bad_alias"}
    runtime.universal_config.cli__plain_output = None

    run_click_command(
        service.service_list, runtime=runtime, plain=False, print_json=False
    )

    cli_tbl.assert_called_with(
        [
            ["Default", "Type", "Name", "Description"],
            [True, "bad", "bad_alias", "Unconfigured Service"],
            [False, "something_else", "", "something else"],
            [True, "test", "test_alias", "Test Service"],
            [False, "test", "test2_alias", "Test Service"],
        ],
        title="Services",
        dim_rows=[2],
    )


@mock.patch("json.dumps")
def test_service_list_json(json_):
    services = {
        "bad": {"description": "Unconfigured Service"},
        "test": {"description": "Test Service"},
    }
    runtime = mock.Mock()
    runtime.project_config.services = services
    runtime.keychain.list_services.return_value = ["test"]
    runtime.universal_config.cli__plain_output = None

    run_click_command(
        service.service_list, runtime=runtime, plain=False, print_json=True
    )

    json_.assert_called_with(services)


def test_service_connect__list_commands():
    multi_cmd = service.ConnectServiceCommand()
    runtime = mock.Mock()
    runtime.project_config.services = {"test": {}}

    with click.Context(multi_cmd, obj=runtime) as ctx:
        result = multi_cmd.list_commands(ctx)
    assert result == ["test"]


def test_service_connect__list_global_keychain():
    multi_cmd = service.ConnectServiceCommand()
    runtime = mock.Mock()
    runtime.project_config = None
    runtime.universal_config.services = {"test": {}}

    with click.Context(multi_cmd, obj=runtime) as ctx:
        result = multi_cmd.list_commands(ctx)
    assert result == ["test"]


def test_service_connect():
    multi_cmd = service.ConnectServiceCommand()
    runtime = mock.MagicMock()
    runtime.keychain.get_default_service_name.return_value = None
    runtime.project_config.services = {
        "test": {"attributes": {"attr": {"required": False}}}
    }

    with click.Context(multi_cmd, obj=runtime) as ctx:
        cmd = multi_cmd.get_command(ctx, "test")
        cmd.callback(ctx.obj, project=True, service_name="test-alias")

        runtime.keychain.set_service.assert_called_once()

        cmd.callback(ctx.obj, project=False, service_name="test-alias")


@mock.patch("click.confirm")
def test_service_connect__alias_already_exists(confirm):
    confirm.side_effect = "y"
    multi_cmd = service.ConnectServiceCommand()
    ctx = mock.Mock()
    runtime = mock.MagicMock()
    runtime.project_config.services = {
        "test-type": {"attributes": {"attr": {"required": False}}}
    }
    runtime.services = {"test-type": {"already-exists": "some config"}}
    runtime.keychain.list_services.return_value = {"test-type": ["already-exists"]}
    runtime.keychain.get_default_service_name.return_value = None

    with click.Context(multi_cmd, obj=runtime) as ctx:
        cmd = multi_cmd.get_command(ctx, "test-type")
        cmd.callback(
            runtime,
            service_type="test-type",
            service_name="already-exists",
            project=True,
        )

    confirm.assert_called_once()


@mock.patch("click.confirm")
def test_service_connect__set_new_service_as_default(confirm):
    confirm.return_value = True
    multi_cmd = service.ConnectServiceCommand()
    runtime = mock.MagicMock()
    runtime.project_config.services = {
        "test-type": {"attributes": {"attr": {"required": False}}}
    }
    service_name = "existing-service"
    runtime.services = {"test-type": {service_name: "some config"}}
    runtime.keychain.list_services.return_value = {"test-type": [service_name]}
    runtime.keychain.get_default_service_name.return_value = service_name

    with click.Context(multi_cmd, obj=runtime) as ctx:
        cmd = multi_cmd.get_command(ctx, "test-type")
        cmd.callback(
            runtime,
            service_type="test-type",
            service_name="new-service",
            project=False,
        )

    confirm.assert_called_once()
    runtime.keychain.set_default_service.assert_called_once_with(
        "test-type", "new-service"
    )


@mock.patch("click.confirm")
def test_service_connect__do_not_set_new_service_as_default(confirm):
    confirm.return_value = False
    multi_cmd = service.ConnectServiceCommand()
    runtime = mock.MagicMock()
    runtime.project_config.services = {
        "test-type": {"attributes": {"attr": {"required": False}}}
    }
    service_name = "existing-service"
    runtime.services = {"test-type": {service_name: "some config"}}
    runtime.keychain.list_services.return_value = {"test-type": [service_name]}
    runtime.keychain.get_default_service_name.return_value = service_name

    with click.Context(multi_cmd, obj=runtime) as ctx:
        cmd = multi_cmd.get_command(ctx, "test-type")
        cmd.callback(
            runtime,
            service_type="test-type",
            service_name="new-service",
            project=False,
        )

    confirm.assert_called_once()
    runtime.keychain.set_default_service.call_count = 0


@mock.patch("click.echo")
def test_service_connect__no_name_given(echo):
    multi_cmd = service.ConnectServiceCommand()
    ctx = mock.Mock()
    runtime = mock.MagicMock()
    runtime.project_config.services = {
        "test-type": {"attributes": {"attr": {"required": False}}}
    }
    runtime.services = {"test-type": {}}
    runtime.keychain.list_services.return_value = {"test-type": []}
    runtime.keychain.get_default_service_name.return_value = None

    with click.Context(multi_cmd, obj=runtime) as ctx:
        cmd = multi_cmd.get_command(ctx, "test-type")
        cmd.callback(
            runtime,
            service_type="test-type",
            service_name=None,
            project=True,
        )

    # service_name is None, so the alias when setting the service should be 'default'
    assert "default" in runtime.keychain.set_service.call_args_list[0][0]
    assert (
        echo.call_args_list[0][0][0]
        == "No service name specified. Using 'default' as the service name."
    )


@mock.patch("click.echo")
def test_service_connect__global_default(echo):
    multi_cmd = service.ConnectServiceCommand()
    ctx = mock.Mock()
    runtime = mock.MagicMock()
    runtime.project_config.services = {
        "test": {"attributes": {"attr": {"required": False}}}
    }
    runtime.keychain.get_default_service_name.return_value = None

    with click.Context(multi_cmd, obj=runtime) as ctx:
        cmd = multi_cmd.get_command(ctx, "test")
        cmd.callback(runtime, service_name="test-alias", default=True, project=False)

    runtime.keychain.set_default_service.assert_called_once_with(
        "test", "test-alias", project=False
    )
    assert echo.call_args_list[0][0][0] == "Service test:test-alias is now connected"
    assert (
        echo.call_args_list[1][0][0]
        == "Service test:test-alias is now the default for all CumulusCI projects"
    )


@mock.patch("click.echo")
def test_service_connect__project_default(echo):
    multi_cmd = service.ConnectServiceCommand()
    ctx = mock.Mock()
    runtime = mock.MagicMock()
    runtime.project_config.services = {
        "test": {"attributes": {"attr": {"required": False}}}
    }
    runtime.keychain.get_default_service_name.return_value = None

    with click.Context(multi_cmd, obj=runtime) as ctx:
        cmd = multi_cmd.get_command(ctx, "test")
        cmd.callback(runtime, service_name="test-alias", default=False, project=True)

    runtime.keychain.set_default_service.assert_called_once_with(
        "test", "test-alias", project=True
    )
    assert echo.call_args_list[0][0][0] == "Service test:test-alias is now connected"
    assert (
        "Service test:test-alias is now the default for project"
        in echo.call_args_list[1][0][0]
    )


def test_service_connect_global_keychain():
    multi_cmd = service.ConnectServiceCommand()
    runtime = mock.MagicMock()
    runtime.project_config = None
    runtime.universal_config.services = {
        "test": {"attributes": {"attr": {"required": False}}}
    }
    runtime.keychain.get_default_service_name.return_value = None

    with click.Context(multi_cmd, obj=runtime) as ctx:
        cmd = multi_cmd.get_command(ctx, "test")
        cmd.callback(ctx.obj, project=True, service_name="test-alias")

        runtime.keychain.set_service.assert_called_once()

        cmd.callback(ctx.obj, project=False, service_name="test-alias")


def test_service_connect_invalid_service():
    multi_cmd = service.ConnectServiceCommand()
    runtime = mock.MagicMock()
    runtime.project_config.services = {}

    with click.Context(multi_cmd, obj=runtime) as ctx:
        with pytest.raises(click.UsageError):
            multi_cmd.get_command(ctx, "test")


@mock.patch("cumulusci.cli.tests.test_service.validate_service")
def test_service_connect_validator(validator):
    multi_cmd = service.ConnectServiceCommand()
    runtime = mock.MagicMock()
    runtime.project_config.services = {
        "test": {
            "attributes": {},
            "validator": "cumulusci.cli.tests.test_service.validate_service",
        }
    }
    runtime.keychain.get_default_service_name.return_value = None

    expected_conf = {
        "service_type": "test",
        "service_name": "test-alias",
        "key": "value",
    }
    validator.return_value = expected_conf
    with click.Context(multi_cmd, obj=runtime) as ctx:
        cmd = multi_cmd.get_command(ctx, "test")
        cmd.callback(
            runtime,
            service_type="test",
            service_name="test-alias",
            project=False,
            key="value",
        )
        validator.assert_called_once_with(expected_conf)


@mock.patch("cumulusci.cli.tests.test_service.validate_service")
def test_service_connect_validator_failure(validator):
    multi_cmd = service.ConnectServiceCommand()
    runtime = mock.MagicMock()
    runtime.project_config.services = {
        "test": {
            "attributes": {},
            "validator": "cumulusci.cli.tests.test_service.validate_service",
        }
    }
    runtime.keychain.get_default_service_name.return_value = None

    validator.side_effect = Exception("Validation failed")
    with click.Context(multi_cmd, obj=runtime) as ctx:
        cmd = multi_cmd.get_command(ctx, "test")
        with pytest.raises(Exception, match="Validation failed"):
            cmd.callback(
                runtime,
                service_type="test",
                service_name="test-alias",
                project=False,
            )


@mock.patch("cumulusci.cli.service.CliTable")
def test_service_info(cli_tbl):
    cli_tbl._table = mock.Mock()
    service_config = mock.Mock()
    service_config.config = {"description": "Test Service"}
    runtime = mock.Mock()
    runtime.keychain.get_service.return_value = service_config
    runtime.universal_config.cli__plain_output = None

    run_click_command(
        service.service_info,
        runtime=runtime,
        service_type="test",
        service_name="test-alias",
        plain=False,
    )

    cli_tbl.assert_called_with(
        [["Key", "Value"], ["\x1b[1mdescription\x1b[0m", "Test Service"]],
        title="test:test-alias",
    )


@mock.patch("click.echo")
def test_service_info_not_configured(echo):
    runtime = mock.Mock()
    runtime.keychain.get_service.side_effect = ServiceNotConfigured

    run_click_command(
        service.service_info,
        runtime=runtime,
        service_type="test",
        service_name="test-alias",
        plain=False,
    )
    assert "not configured for this project" in echo.call_args[0][0]


@mock.patch("click.echo")
def test_service_default__global(echo):
    runtime = mock.Mock()
    run_click_command(
        service.service_default,
        runtime=runtime,
        service_type="test",
        service_name="test-alias",
        project=False,
    )
    runtime.keychain.set_default_service.called_once_with("test", "test-alias")
    echo.assert_called_once_with(
        "Service test:test-alias is now the default for all CumulusCI projects"
    )


@mock.patch("click.echo")
def test_service_default__project(echo):
    runtime = mock.Mock()
    runtime.keychain.project_local_dir = "test"
    run_click_command(
        service.service_default,
        runtime=runtime,
        service_type="test",
        service_name="test-alias",
        project=True,
    )
    runtime.keychain.set_default_service.called_once_with("test", "test-alias")
    echo.assert_called_once_with(
        "Service test:test-alias is now the default for project 'test'"
    )


@mock.patch("click.echo")
def test_service_default__exception(echo):
    runtime = mock.Mock()
    runtime.keychain.set_default_service.side_effect = ServiceNotConfigured(
        "test error"
    )
    run_click_command(
        service.service_default,
        runtime=runtime,
        service_type="no-such-type",
        service_name="test-alias",
        project=False,
    )
    echo.assert_called_once_with(
        "An error occurred setting the default service: test error"
    )


@mock.patch("click.echo")
def test_service_rename(echo):
    runtime = mock.Mock()
    run_click_command(
        service.service_rename,
        runtime=runtime,
        service_type="test-type",
        current_name="old-alias",
        new_name="new-alias",
    )
    runtime.keychain.rename_service.assert_called_once_with(
        "test-type", "old-alias", "new-alias"
    )
    echo.assert_called_once_with(
        "Service test-type:old-alias has been renamed to new-alias"
    )


@mock.patch("click.echo")
def test_service_rename__exception(echo):
    runtime = mock.Mock()
    runtime.keychain.rename_service.side_effect = ServiceNotConfigured("test error")
    run_click_command(
        service.service_rename,
        runtime=runtime,
        service_type="test-type",
        current_name="old-alias",
        new_name="new-alias",
    )
    echo.assert_called_once_with("An error occurred renaming the service: test error")


@mock.patch("cumulusci.cli.service.click")
def test_service_remove(click):
    click.prompt.side_effect = ("future-default-alias",)
    runtime = mock.Mock()
    runtime.keychain.env_service_var_prefix = "CUMULUSCI_SERVICE_"
    runtime.keychain.services = {
        "github": {
            "current-default-alias": "config1",
            "another-alias": "config2",
            "future-default-alias": "config3",
        }
    }
    runtime.keychain._default_services = {"github": "current-default-alias"}
    runtime.keychain.list_services.return_value = {
        "github": ["current-default-alias", "another-alias", "future-default-alias"]
    }
    run_click_command(
        service.service_remove,
        runtime=runtime,
        service_type="github",
        service_name="current-default-alias",
    )
    runtime.keychain.remove_service.assert_called_once_with(
        "github", "current-default-alias"
    )
    runtime.keychain.set_default_service.assert_called_once_with(
        "github", "future-default-alias"
    )
    assert (
        click.echo.call_args_list[-1][0][0]
        == "Service github:current-default-alias has been removed."
    )


@mock.patch("cumulusci.cli.service.click")
def test_service_remove__name_does_not_exist(click):
    click.prompt.side_effect = ("this-alias-does-not-exist",)
    runtime = mock.Mock()
    runtime.keychain.env_service_var_prefix = "CUMULUSCI_SERVICE_"
    runtime.keychain.services = {
        "github": {
            "current-default-alias": "config1",
            "another-alias": "config2",
            "future-default-alias": "config3",
        }
    }
    runtime.keychain._default_services = {"github": "current-default-alias"}
    runtime.keychain.list_services.return_value = {
        "github": ["current-default-alias", "another-alias", "future-default-alias"]
    }
    run_click_command(
        service.service_remove,
        runtime=runtime,
        service_type="github",
        service_name="current-default-alias",
    )
    assert (
        click.echo.call_args_list[-1][0][0]
        == "No service of type github with name: this-alias-does-not-exist"
    )
    assert runtime.keychain.remove_service.call_count == 0
    assert runtime.keychain.set_default_service.call_count == 0


@mock.patch("cumulusci.cli.service.click")
def test_service_remove__exception_thrown(click):
    click.prompt.side_effect = ("future-default-alias",)
    runtime = mock.Mock()
    runtime.keychain.env_service_var_prefix = "CUMULUSCI_SERVICE_"
    runtime.keychain.services = {
        "github": {
            "current-default-alias": "config1",
            "another-alias": "config2",
            "future-default-alias": "config3",
        }
    }
    runtime.keychain._default_services = {"github": "current-default-alias"}
    runtime.keychain.list_services.return_value = {
        "github": ["current-default-alias", "another-alias", "future-default-alias"]
    }
    runtime.keychain.remove_service.side_effect = ServiceNotConfigured("test error")
    run_click_command(
        service.service_remove,
        runtime=runtime,
        service_type="github",
        service_name="current-default-alias",
    )
    assert (
        click.echo.call_args_list[-1][0][0]
        == "An error occurred removing the service: test error"
    )


@mock.patch("cumulusci.cli.service.click")
def test_service_remove__environment_service_cannot_be_removed(click):
    runtime = mock.Mock()
    runtime.keychain.env_service_var_prefix = "CUMULUSCI_SERVICE_"
    runtime.keychain.services = {
        "github": {
            "env-foo": "config-from-env",
            "another-alias": "config-from-file",
        }
    }
    with EnvironmentVarGuard() as env:
        env.set(
            "CUMULUSCI_SERVICE_github__env-foo", '{"username":"foo", "token": "bar"}'
        )
        run_click_command(
            service.service_remove,
            runtime=runtime,
            service_type="github",
            service_name="env-foo",
        )
    assert (
        click.echo.call_args_list[-1][0][0]
        == "The service github:env-foo is defined by environment variables. If you would like it removed please delete the environment variable with name: CUMULUSCI_SERVICE_github__env-foo"
    )


def validate_service(options):
    pass



import base64
import enum
import io
import json
import pathlib
import zipfile
from typing import List, Optional

from pydantic import BaseModel, validator
from simple_salesforce.exceptions import SalesforceMalformedRequest

from cumulusci.core.config.util import get_devhub_config
from cumulusci.core.dependencies.dependencies import (
    PackageNamespaceVersionDependency,
    PackageVersionIdDependency,
    UnmanagedDependency,
    UnmanagedGitHubRefDependency,
)
from cumulusci.core.dependencies.resolvers import get_static_dependencies
from cumulusci.core.exceptions import (
    CumulusCIUsageError,
    DependencyLookupError,
    GithubException,
    PackageUploadFailure,
    TaskOptionsError,
)
from cumulusci.core.github import get_version_id_from_tag
from cumulusci.core.sfdx import convert_sfdx_source
from cumulusci.core.utils import process_bool_arg
from cumulusci.core.versions import PackageVersionNumber, VersionTypeEnum
from cumulusci.salesforce_api.package_zip import (
    BasePackageZipBuilder,
    MetadataPackageZipBuilder,
)
from cumulusci.salesforce_api.utils import get_simple_salesforce_connection
from cumulusci.tasks.salesforce.BaseSalesforceApiTask import BaseSalesforceApiTask
from cumulusci.tasks.salesforce.org_settings import build_settings_package
from cumulusci.utils.git import split_repo_url


class PackageTypeEnum(str, enum.Enum):
    managed = "Managed"
    unlocked = "Unlocked"


class PackageConfig(BaseModel):
    package_name: str
    description: str = ""
    package_type: PackageTypeEnum
    org_dependent: bool = False
    post_install_script: Optional[str]
    uninstall_script: Optional[str]
    namespace: Optional[str]
    version_name: str
    version_base: Optional[str]
    version_type: VersionTypeEnum = VersionTypeEnum.minor

    @validator("org_dependent")
    def org_dependent_must_be_unlocked(cls, v, values):
        if v and values["package_type"] != PackageTypeEnum.unlocked:
            raise ValueError("Only unlocked packages can be org-dependent.")
        return v

    @validator("post_install_script")
    def post_install_script_must_be_managed(cls, v, values):
        if v and values["package_type"] != PackageTypeEnum.managed:
            raise ValueError("Only managed packages can have a post-install script.")
        return v

    @validator("uninstall_script")
    def uninstall_script_must_be_managed(cls, v, values):
        if v and values["package_type"] != PackageTypeEnum.managed:
            raise ValueError("Only managed packages can have an uninstall script.")
        return v


class CreatePackageVersion(BaseSalesforceApiTask):
    """Creates a new second-generation package version.

    If a package named ``package_name`` does not yet exist in the Dev Hub, it will be created.
    """

    api_version = "52.0"

    task_options = {
        "package_name": {"description": "Name of package"},
        "package_type": {
            "description": "Package type (Unlocked or Managed)",
            "required": True,
        },
        "namespace": {"description": "Package namespace"},
        "version_name": {"description": "Version name"},
        "version_base": {
            "description": "The version number to use as a base before incrementing. "
            "Optional; defaults to the highest existing version number of this package. "
            "Can be set to ``latest_github_release`` to use the version of the most recent release published to GitHub."
        },
        "version_type": {
            "description": "The part of the version number to increment. "
            "Options are major, minor, patch, build.  Defaults to build"
        },
        "skip_validation": {
            "description": "If true, skip validation of the package version. Default: false. "
            "Skipping validation creates packages more quickly, but they cannot be promoted for release."
        },
        "org_dependent": {
            "description": "If true, create an org-dependent unlocked package. Default: false."
        },
        "post_install_script": {
            "description": "Post-install script (for managed packages)",
        },
        "uninstall_script": {
            "description": "Uninstall script (for managed packages)",
        },
        "force_upload": {
            "description": "If true, force creating a new package version even if one with the same contents already exists"
        },
        "static_resource_path": {
            "description": "The path where decompressed static resources are stored. "
            "Any subdirectories found will be zipped and added to the staticresources directory of the build."
        },
        "ancestor_id": {
            "description": "The 04t Id to use for the ancestor of this package. "
            "Optional; defaults to no ancestor specified. "
            "Can be set to ``latest_github_release`` to use the most recent production version published to GitHub."
        },
        "resolution_strategy": {
            "description": "The name of a sequence of resolution_strategy "
            "(from project__dependency_resolutions) to apply to dynamic dependencies. Defaults to 'production'."
        },
        "create_unlocked_dependency_packages": {
            "description": "If True, create unlocked packages for unpackaged metadata in this project and dependencies. "
            "Defaults to False."
        },
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)

        self.package_config = PackageConfig(
            package_name=self.options.get("package_name")
            or self.project_config.project__package__name,
            package_type=self.options.get("package_type")
            or self.project_config.project__package__type,
            org_dependent=self.options.get("org_dependent", False),
            post_install_script=self.options.get("post_install_script")
            or self.project_config.project__package__install_class,
            uninstall_script=self.options.get("uninstall_script")
            or self.project_config.project__package__uninstall_class,
            namespace=self.options.get("namespace")
            or self.project_config.project__package__namespace,
            version_name=self.options.get("version_name") or "Release",
            version_base=self.options.get("version_base"),
            version_type=self.options.get("version_type") or "build",
        )
        self.options["skip_validation"] = process_bool_arg(
            self.options.get("skip_validation") or False
        )
        self.options["force_upload"] = process_bool_arg(
            self.options.get("force_upload") or False
        )
        self.options["create_unlocked_dependency_packages"] = process_bool_arg(
            self.options.get("create_unlocked_dependency_packages") or False
        )

    def _init_task(self):
        self.tooling = get_simple_salesforce_connection(
            self.project_config,
            get_devhub_config(self.project_config),
            api_version=self.api_version,
            base_url="tooling",
        )

    def _run_task(self):
        """Creates a new 2GP package version.

        1. Create package if not found in Dev Hub.
        2. Request creation of package version.
        3. Wait for completion.
        4. Collect package information as return values.
        """
        # find existing package in Dev Hub, or create one if necessary
        self.package_id = self._get_or_create_package(self.package_config)
        self.return_values["package_id"] = self.package_id

        # submit request to create package version
        options = {
            "package_type": self.package_config.package_type.value,
            "namespace_inject": self.package_config.namespace,
            "namespaced_org": self.package_config.namespace is not None,
        }
        if "static_resource_path" in self.options:
            options["static_resource_path"] = self.options["static_resource_path"]

        package_zip_builder = None
        with convert_sfdx_source(
            self.project_config.default_package_path,
            self.package_config.package_name,
            self.logger,
        ) as path:
            package_zip_builder = MetadataPackageZipBuilder(
                path=path,
                name=self.package_config.package_name,
                options=options,
                logger=self.logger,
            )

        ancestor_id = self._resolve_ancestor_id(self.options.get("ancestor_id"))

        self.request_id = self._create_version_request(
            self.package_id,
            self.package_config,
            package_zip_builder,
            ancestor_id,
            self.options["skip_validation"],
        )
        self.return_values["request_id"] = self.request_id

        # wait for request to complete
        self._poll()
        self.return_values["package2_version_id"] = self.package_version_id

        # get the new version number from Package2Version
        res = self.tooling.query(
            f"SELECT MajorVersion, MinorVersion, PatchVersion, BuildNumber, SubscriberPackageVersionId FROM Package2Version WHERE Id='{self.package_version_id}'"
        )
        package2_version = res["records"][0]
        self.return_values["subscriber_package_version_id"] = package2_version[
            "SubscriberPackageVersionId"
        ]
        self.return_values["version_number"] = PackageVersionNumber(
            **package2_version
        ).format()

        # get the new version's dependencies from SubscriberPackageVersion
        res = self.tooling.query(
            "SELECT Dependencies FROM SubscriberPackageVersion "
            f"WHERE Id='{package2_version['SubscriberPackageVersionId']}'"
        )
        self.return_values["dependencies"] = self._prepare_cci_dependencies(
            res["records"][0]["Dependencies"]
        )

        self.logger.info("Created package version:")
        self.logger.info(f"  Package2 Id: {self.package_id}")
        self.logger.info(f"  Package2Version Id: {self.package_version_id}")
        self.logger.info(
            f"  SubscriberPackageVersion Id: {self.return_values['subscriber_package_version_id']}"
        )
        self.logger.info(f"  Version Number: {self.return_values['version_number']}")
        self.logger.info(f"  Dependencies: {self.return_values['dependencies']}")

    def _get_or_create_package(self, package_config: PackageConfig):
        """Find or create the Package2

        Checks the Dev Hub for an existing, non-deprecated 2GP package
        with matching name, type, and namespace.
        """
        message = f"Checking for existing {package_config.package_type} Package named {package_config.package_name}"
        query = (
            f"SELECT Id, ContainerOptions FROM Package2 WHERE IsDeprecated = FALSE "
            f"AND ContainerOptions='{package_config.package_type}' "
            f"AND IsOrgDependent={package_config.org_dependent} "
            f"AND Name='{package_config.package_name}'"
        )
        if package_config.namespace:
            query += f" AND NamespacePrefix='{package_config.namespace}'"
            message += f" with namespace {package_config.namespace}"
        else:
            query += " AND NamespacePrefix=null"
        self.logger.info(message)
        try:
            res = self.tooling.query(query)
        except SalesforceMalformedRequest as err:
            if "Object type 'Package2' is not supported" in err.content[0]["message"]:
                raise TaskOptionsError(
                    "This org does not have a Dev Hub with 2nd-generation packaging enabled. "
                    "Make sure you are using the correct org and/or check the Dev Hub settings in Setup."
                )
            raise  # pragma: no cover
        if res["size"] > 1:
            raise TaskOptionsError(
                f"Found {res['size']} packages with the same name, namespace, and package_type"
            )
        if res["size"] == 1:
            existing_package = res["records"][0]
            if existing_package["ContainerOptions"] != package_config.package_type:
                raise PackageUploadFailure(
                    f"Duplicate Package: {existing_package['ContainerOptions']} package with id "
                    f"{ existing_package['Id']} has the same name ({package_config.package_name}) "
                    "for this namespace but has a different package type"
                )
            package_id = existing_package["Id"]
            self.logger.info(f"Found {package_id}")
            return package_id

        self.logger.info("No existing package found, creating the package")
        Package2 = self._get_tooling_object("Package2")
        package = Package2.create(
            {
                "ContainerOptions": package_config.package_type,
                "IsOrgDependent": package_config.org_dependent,
                "Name": package_config.package_name,
                "Description": package_config.description,
                "NamespacePrefix": package_config.namespace,
            }
        )
        return package["id"]

    def _create_version_request(
        self,
        package_id: str,
        package_config: PackageConfig,
        package_zip_builder: BasePackageZipBuilder,
        ancestor_id: str = "",
        skip_validation: bool = False,
        dependencies: list = None,
    ):
        # Prepare the VersionInfo file
        version_bytes = io.BytesIO()
        version_info = zipfile.ZipFile(version_bytes, "w", zipfile.ZIP_DEFLATED)
        try:

            # Add the package.zip
            package_hash = package_zip_builder.as_hash()
            version_info.writestr("package.zip", package_zip_builder.as_bytes())

            if not self.options["force_upload"]:
                # Check for an existing package with the same contents
                res = self.tooling.query(
                    "SELECT Id "
                    "FROM Package2VersionCreateRequest "
                    f"WHERE Package2Id = '{package_id}' "
                    "AND Status != 'Error' "
                    f"AND SkipValidation = {str(skip_validation)} "
                    f"AND Tag = 'hash:{package_hash}' "
                    "ORDER BY CreatedDate DESC"
                )
                if res["size"] > 0:
                    self.logger.info(
                        "Found existing request for package with the same metadata.  Using existing package."
                    )
                    return res["records"][0]["Id"]

            # Create the package descriptor
            version_number = self._get_base_version_number(
                package_config.version_base, package_id
            ).increment(package_config.version_type)

            package_descriptor = {
                "id": package_id,
                "path": "",
                "versionName": package_config.version_name,
                "versionNumber": version_number.format(),
                "ancestorId": ancestor_id,
            }

            if package_config.post_install_script:
                package_descriptor[
                    "postInstallScript"
                ] = package_config.post_install_script
            if package_config.uninstall_script:
                package_descriptor["uninstallScript"] = package_config.uninstall_script

            # Add org shape
            with open(self.org_config.config_file, "r") as f:
                scratch_org_def = json.load(f)
            for key in (
                "country",
                "edition",
                "language",
                "features",
                "snapshot",
            ):
                if key in scratch_org_def:
                    package_descriptor[key] = scratch_org_def[key]

            # Add settings
            if "settings" in scratch_org_def or "objectSettings" in scratch_org_def:
                with build_settings_package(
                    scratch_org_def.get("settings"),
                    scratch_org_def.get("objectSettings"),
                    self.api_version,
                ) as path:
                    settings_zip_builder = MetadataPackageZipBuilder(path=path)
                    version_info.writestr(
                        "settings.zip", settings_zip_builder.as_bytes()
                    )

            # Add the dependencies for the package
            is_dependency = package_config is not self.package_config
            if (
                not (package_config.org_dependent or skip_validation)
                and not is_dependency
            ):
                self.logger.info("Determining dependencies for package")
                dependencies = self._get_dependencies()
            if dependencies:
                package_descriptor["dependencies"] = dependencies

            # Add package descriptor to version info
            version_info.writestr(
                "package2-descriptor.json", json.dumps(package_descriptor)
            )
        finally:
            version_info.close()
        version_info = base64.b64encode(version_bytes.getvalue()).decode("utf-8")
        Package2CreateVersionRequest = self._get_tooling_object(
            "Package2VersionCreateRequest"
        )
        request = {
            "Package2Id": package_id,
            "SkipValidation": skip_validation,
            "Tag": f"hash:{package_hash}",
            "VersionInfo": version_info,
            "CalculateCodeCoverage": not skip_validation,
        }
        self.logger.info(
            f"Requesting creation of package version {version_number.format()} "
            f"for package {package_config.package_name} ({package_id})"
        )
        response = Package2CreateVersionRequest.create(request)
        self.logger.info(
            f"Package2VersionCreateRequest created with id {response['id']}"
        )
        return response["id"]

    def _resolve_ancestor_id(self, spv_id: str = None) -> str:
        """
        If an ancestor_id (04t) is not specified, get it
        from the latest production release.

        @param spv_id The SubscriberPackageVersionId (04t) that is the ancestor
        to the version being created.
        """
        if not spv_id:
            return ""
        elif self.package_config.package_type == PackageTypeEnum.unlocked:
            raise CumulusCIUsageError(
                "Cannot specify an ancestor for Unlocked packages."
            )
        elif spv_id.startswith("04t"):
            return self._convert_ancestor_id(spv_id)
        elif spv_id == "latest_github_release":
            try:
                tag_name = self.project_config.get_latest_tag(beta=False)
            except GithubException:
                # No release found
                return ""
            repo = self.project_config.get_repo()
            spv_id = get_version_id_from_tag(repo, tag_name)
            self.logger.info(f"Resolved ancestor to version: {spv_id}")
            self.logger.info("")

            return self._convert_ancestor_id(spv_id)
        else:
            raise TaskOptionsError(f"Unrecognized value for ancestor_id: {spv_id}")

    def _convert_ancestor_id(self, ancestor_id: str) -> str:
        """Given a SubscriberPackageVersionId (04t) find
        the corresponding Package2VersionId (05i).
        See: https://github.com/forcedotcom/salesforce-alm/blob/83745351670a701762c6ecc926885564b8853357/src/lib/package/packageUtils.ts#L517

        @param ancestor_id A SubscriberPackageVersionId (04t)
        @returns the corresponding Package2VersionId (05i) or an empty string
        if no Package2Version is found.
        """
        package_2_version_id = ""
        res = self.tooling.query(
            f"SELECT Id FROM Package2Version WHERE SubscriberPackageVersionId='{ancestor_id}'"
        )
        if res["size"] > 0:
            package_2_version_id = res["records"][0]["Id"]
            self.logger.info(
                f"Converted ancestor_id to corresponding Package2Version: {package_2_version_id}"
            )
            self.logger.info("")

        return package_2_version_id

    def _get_base_version_number(
        self, version_base: Optional[str], package_id: str
    ) -> PackageVersionNumber:
        """Determine the "base version" of the package (existing version to be incremented)"""
        if version_base is None:
            # Default: Get the highest existing version of the package
            res = self.tooling.query(
                "SELECT MajorVersion, MinorVersion, PatchVersion, BuildNumber, IsReleased "
                "FROM Package2Version "
                f"WHERE Package2Id='{package_id}' "
                "ORDER BY MajorVersion DESC, MinorVersion DESC, PatchVersion DESC, BuildNumber DESC "
                "LIMIT 1"
            )
            if res["size"]:
                return PackageVersionNumber(**res["records"][0])
        elif version_base == "latest_github_release":
            # Get the version of the latest github release
            try:
                return PackageVersionNumber.parse(
                    str(self.project_config.get_latest_version())
                )
            except GithubException:
                # handle case where there isn't a release yet
                pass
        else:
            return PackageVersionNumber.parse(version_base)
        return PackageVersionNumber()

    def _get_dependencies(self):
        """Resolve dependencies into SubscriberPackageVersionIds (04t prefix)"""
        dependencies = get_static_dependencies(
            self.project_config,
            resolution_strategy=self.options.get("resolution_strategy") or "production",
        )

        # If any dependencies are expressed as a 1gp namespace + version,
        # we need to convert those to 04t package version ids,
        # for which we need an org with the packages installed.
        if self._has_1gp_namespace_dependency(dependencies):
            dependencies = self.org_config.resolve_04t_dependencies(dependencies)

        # Convert dependencies to correct format for Package2VersionCreateRequest
        dependencies = self._convert_project_dependencies(dependencies)

        # Build additional packages for local unpackaged/pre
        dependencies = self._get_unpackaged_pre_dependencies(dependencies)

        return dependencies

    def _has_1gp_namespace_dependency(self, project_dependencies):
        """Returns true if any dependencies are specified using a namespace rather than 04t"""
        return any(
            isinstance(dependency, PackageNamespaceVersionDependency)
            and not dependency.version_id
            for dependency in project_dependencies
        )

    def _convert_project_dependencies(self, dependencies):
        """Convert dependencies into the format expected by Package2VersionCreateRequest.

        For dependencies expressed as a GitHub repo subfolder, build an unlocked package from that.
        """
        new_dependencies = []
        for dependency in dependencies:
            new_dependency = {}
            if isinstance(dependency, PackageVersionIdDependency) or (
                isinstance(dependency, PackageNamespaceVersionDependency)
                and dependency.version_id
            ):
                self.logger.info(
                    f"Adding dependency {dependency.package_name} with id {dependency.version_id}"
                )
                new_dependency["subscriberPackageVersionId"] = dependency.version_id
            elif isinstance(dependency, UnmanagedDependency):
                if self.options["create_unlocked_dependency_packages"]:
                    version_id = self._create_unlocked_package_from_unmanaged_dep(
                        dependency, new_dependencies
                    )
                    self.logger.info(
                        f"Adding dependency {dependency} with id {version_id}"
                    )
                    new_dependency["subscriberPackageVersionId"] = version_id
                else:
                    self.logger.info(
                        f"Skipping dependency {dependency} because create_unlocked_dependency_packages is False."
                    )
                    continue
            else:
                raise DependencyLookupError(
                    f"Unable to convert dependency: {dependency}"
                )

            new_dependencies.append(new_dependency)

        return new_dependencies

    def _get_unpackaged_pre_dependencies(self, dependencies):
        """Create package for unpackaged/pre metadata, if necessary"""
        path = pathlib.Path("unpackaged", "pre")
        if not path.exists():
            return dependencies

        if not self.options["create_unlocked_dependency_packages"]:
            self.logger.info(
                "Skipping unpackaged/pre dependencies because create_unlocked_dependency_packages is False."
            )
            return dependencies

        for item_path in sorted(path.iterdir(), key=str):
            if item_path.is_dir():
                version_id = self._create_unlocked_package_from_local(
                    item_path, dependencies
                )
                self.logger.info(
                    "Adding dependency {}/{} {} with id {}".format(
                        self.project_config.repo_owner,
                        self.project_config.repo_name,
                        item_path,
                        version_id,
                    )
                )
                dependencies.append({"subscriberPackageVersionId": version_id})

        return dependencies

    def _create_unlocked_package_from_unmanaged_dep(
        self, dependency: UnmanagedDependency, dependencies
    ) -> str:
        if isinstance(dependency, UnmanagedGitHubRefDependency):
            repo_owner, repo_name = split_repo_url(dependency.github)
            package_name = f"{repo_owner}/{repo_name} {dependency.subfolder}"
        else:
            package_name = dependency.description

        package_zip_builder = dependency.get_metadata_package_zip_builder(
            self.project_config, self.org_config
        )

        package_config = PackageConfig(
            package_name=package_name,
            version_name="Auto",
            package_type="Unlocked",
            # Ideally we'd do this without a namespace,
            # but it needs to match the dependent package
            namespace=self.package_config.namespace,
        )
        package_id = self._get_or_create_package(package_config)
        self.request_id = self._create_version_request(
            package_id,
            package_config,
            package_zip_builder,
            dependencies=dependencies,
        )

        self._poll()
        self._reset_poll()
        res = self.tooling.query(
            "SELECT SubscriberPackageVersionId FROM Package2Version "
            f"WHERE Id='{self.package_version_id}'"
        )
        package2_version = res["records"][0]
        return package2_version["SubscriberPackageVersionId"]

    def _create_unlocked_package_from_local(self, path, dependencies):
        """Create an unlocked package version from a local directory."""
        self.logger.info(f"Creating package for dependencies in {path}")
        package_name = (
            f"{self.project_config.repo_owner}/{self.project_config.repo_name} {path}"
        )
        with convert_sfdx_source(path, package_name, self.logger) as src_path:
            package_zip_builder = MetadataPackageZipBuilder(
                path=src_path, name=package_name, logger=self.logger
            )
            package_config = PackageConfig(
                package_name=package_name,
                version_name="Auto",
                package_type="Unlocked",
                # Ideally we'd do this without a namespace,
                # but it needs to match the dependent package
                namespace=self.package_config.namespace,
            )
            package_id = self._get_or_create_package(package_config)
            self.request_id = self._create_version_request(
                package_id,
                package_config,
                package_zip_builder,
                dependencies=dependencies,
            )
        self._poll()
        self._reset_poll()
        res = self.tooling.query(
            "SELECT SubscriberPackageVersionId FROM Package2Version "
            f"WHERE Id='{self.package_version_id}'"
        )
        package2_version = res["records"][0]
        return package2_version["SubscriberPackageVersionId"]

    def _poll_action(self):
        """Check if Package2VersionCreateRequest has completed."""
        res = self.tooling.query(
            f"SELECT Id, Status, Package2VersionId FROM Package2VersionCreateRequest WHERE Id = '{self.request_id}'"
        )
        request = res["records"][0]
        if request["Status"] == "Success":
            self.logger.info("[Success]: Package creation successful")
            self.poll_complete = True
            self.package_version_id = request["Package2VersionId"]
        elif request["Status"] == "Error":
            self.logger.error("[Error]: Package creation failed with error:")
            res = self.tooling.query(
                "SELECT Message FROM Package2VersionCreateRequestError "
                f"WHERE ParentRequestId = '{request['Id']}'"
            )
            errors = []
            if res["size"] > 0:
                for error in res["records"]:
                    errors.append(error["Message"])
                    self.logger.error(error["Message"])
            raise PackageUploadFailure("\n".join(errors))
        else:
            self.logger.info(f"[{request['Status']}]")

    def _prepare_cci_dependencies(self, deps) -> List[dict]:
        # Convert the dependencies returned by the Tooling API
        # for the new package back into `update_dependencies`-compatible
        # format for persistence into the GitHub release.

        if deps:
            return [
                PackageVersionIdDependency(
                    version_id=v["subscriberPackageVersionId"]
                ).dict(exclude_none=True)
                for v in deps["ids"]
            ]

        return []



from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.utils import process_list_arg
from cumulusci.tasks.salesforce import BaseSalesforceApiTask


class ActivateFlow(BaseSalesforceApiTask):
    """
    Activate the Flows with the supplied Developer Names
    """

    task_options = {
        "developer_names": {
            "description": "List of DeveloperNames to query in SOQL",
            "required": True,
        }
    }

    def _init_options(self, kwargs):
        super(ActivateFlow, self)._init_options(kwargs)
        self.options["developer_names"] = process_list_arg(
            self.options.get("developer_names")
        )
        self.api_version = "43.0"
        if not self.options["developer_names"]:
            raise TaskOptionsError(
                "Error you are missing developer_names definition in your task cumulusci.yml file. Please pass in developer_names for your task configuration or use -o to developer_names as a commandline argument"
            )

    def _run_task(self):
        self.logger.info(
            f"Activating the following Flows: {self.options['developer_names']}"
        )
        self.logger.info("Querying flow definitions...")
        result = self.tooling.query(
            "SELECT Id, ActiveVersion.VersionNumber, LatestVersion.VersionNumber, DeveloperName FROM FlowDefinition WHERE DeveloperName IN ({0})".format(
                ",".join([f"'{n}'" for n in self.options["developer_names"]])
            )
        )
        results = []
        for listed_flow in result["records"]:
            results.append(listed_flow["DeveloperName"])
            self.logger.info(f'Processing: {listed_flow["DeveloperName"]}')
            path = f"tooling/sobjects/FlowDefinition/{listed_flow['Id']}"
            urlpath = self.sf.base_url + path
            data = {
                "Metadata": {
                    "activeVersionNumber": listed_flow["LatestVersion"]["VersionNumber"]
                }
            }
            response = self.tooling._call_salesforce("PATCH", urlpath, json=data)
            self.logger.info(response)
        excluded = []
        for i in self.options["developer_names"]:
            if i not in results:
                excluded.append(i)
        if len(excluded) > 0:
            self.logger.warning(
                f"The following developer names were not found: {excluded}"
            )



from rst2ansi import rst2ansi

from cumulusci.tasks.salesforce import BaseSalesforceApiTask


class ListCommunities(BaseSalesforceApiTask):
    api_version = "46.0"
    task_docs = """
    Lists Communities for the current org via the Connect API.
    """
    task_options = {}

    def _run_task(self):
        communities = self.sf.restful("connect/communities")["communities"]

        nameString = "\n==========================================\n"

        communities_output = ["The current Communities in the org are:\n"]
        for community in communities:
            communities_output.append("\n{}{}".format(community["name"], nameString))
            communities_output.append("* **Id:** {}\n".format(community["id"]))
            communities_output.append("* **Status:** {}\n".format(community["status"]))
            communities_output.append(
                "* **Site Url:** `<{}>`_\n".format(community["siteUrl"])
            )
            communities_output.append(
                "* **Url Path Prefix:** {}\n".format(
                    community.get("urlPathPrefix") or ""
                )
            )
            communities_output.append(
                "* **Template:** {}\n".format(community["templateName"])
            )
            communities_output.append(
                "* **Description:** {}\n".format(community["description"])
            )

        communities_output2 = "\n".join(communities_output).encode("utf-8")
        self.logger.info(rst2ansi(communities_output2))
        self.return_values = [community["name"] for community in communities]



import json
import sys

import pytest
import yaml

from cumulusci.utils.yaml import cumulusci_yml


class TestSchema:
    @pytest.mark.skipif(
        sys.version_info < (3, 7), reason="requires python3.7 or higher"
    )
    def test_schema_validates(self, cumulusci_test_repo_root):
        from jsonschema import validate

        schemapath = (
            cumulusci_test_repo_root / "cumulusci/schema/cumulusci.jsonschema.json"
        )
        with open(schemapath) as f:
            schema = json.load(f)
        with open(cumulusci_test_repo_root / "cumulusci.yml") as f:
            data = yaml.safe_load(f)

        assert validate(data, schema=schema) is None

    def test_schema_is_current(self, cumulusci_test_repo_root):
        current_schema = cumulusci_yml.CumulusCIRoot.schema()
        schemapath = (
            cumulusci_test_repo_root / "cumulusci/schema/cumulusci.jsonschema.json"
        )
        with open(schemapath) as f:
            saved_schema = json.load(f)

        assert current_schema == saved_schema, (current_schema, saved_schema)



import json
import logging
import shutil
import typing as T
from contextlib import contextmanager
from multiprocessing import Queue
from pathlib import Path
from traceback import format_exc

from pydantic import BaseModel

from cumulusci.core.config import (
    BaseConfig,
    BaseProjectConfig,
    ConnectedAppOAuthConfig,
    OrgConfig,
    TaskConfig,
    UniversalConfig,
)
from cumulusci.core.exceptions import ServiceNotConfigured
from cumulusci.core.utils import import_global


class SharedConfig(BaseModel):
    "Properties available both in the Queue and also each worker"
    task_class: type
    project_config: BaseProjectConfig
    org_config: OrgConfig
    failures_dir: Path
    redirect_logging: bool
    connected_app: T.Optional[BaseConfig]  # send a list of services
    outbox_dir: Path  # where do jobs go when they are done

    class Config:
        arbitrary_types_allowed = True


class WorkerConfig(SharedConfig):
    working_dir: Path
    task_options: T.Mapping

    def as_dict(self):
        """Convert to a dict of basic data structures/types, similar to JSON."""
        # runs in the parent process
        return {
            "task_class": dotted_class_name(self.task_class),
            "org_config_class": dotted_class_name(self.org_config.__class__),
            "task_options": self.task_options,
            "working_dir": str(self.working_dir),
            "outbox_dir": str(self.outbox_dir),
            "failures_dir": str(self.failures_dir),
            "org_config": (
                self.org_config.config,
                self.org_config.name,
            ),
            "connected_app": self.connected_app.config if self.connected_app else None,
            "redirect_logging": self.redirect_logging,
            "project_config": {
                "project": {"package": self.project_config.config["project"]["package"]}
            },
        }

    @staticmethod
    def from_dict(worker_config_json):  # todo: rename to `worker_config_dct`
        """Read from a dict of basic data structures/types, similar to JSON."""
        # runs in the child process
        org_config_class = import_global(worker_config_json["org_config_class"])
        org_config = org_config_class(*worker_config_json["org_config"])

        task_options = worker_config_json["task_options"]

        universal_config = UniversalConfig()
        project_config = BaseProjectConfig(
            universal_config,
            config=worker_config_json["project_config"],
        )
        return WorkerConfig(
            task_class=import_global(worker_config_json["task_class"]),
            task_options=task_options,
            project_config=project_config,
            org_config=org_config,
            working_dir=Path(worker_config_json["working_dir"]),
            outbox_dir=Path(worker_config_json["outbox_dir"]),
            failures_dir=Path(worker_config_json["failures_dir"]),
            connected_app=ConnectedAppOAuthConfig(worker_config_json["connected_app"])
            if worker_config_json["connected_app"]
            else None,
            redirect_logging=worker_config_json["redirect_logging"],
        )


def dotted_class_name(cls):
    """Generate a dotted class name for a class object"""
    return cls.__module__ + "." + cls.__name__


class TaskWorker:
    """This class runs in a sub-thread or sub-process"""

    def __init__(self, worker_dict, results_reporter):
        self.worker_config = WorkerConfig.from_dict(worker_dict)
        self.redirect_logging = worker_dict["redirect_logging"]
        self.results_reporter = results_reporter

    def __getattr__(self, name):
        """Easy access to names from the config"""
        return getattr(self.worker_config, name)

    def _make_task(self, task_class, logger):
        """Instantiate a CCI task"""
        if "working_directory" in self.task_class.task_options:
            self.task_options["working_directory"] = self.worker_config.working_dir
        task_config = TaskConfig({"options": self.task_options})
        connected_app = self.connected_app
        keychain = SubprocessKeyChain(connected_app)
        self.project_config.set_keychain(keychain)
        self.org_config.keychain = keychain
        return task_class(
            project_config=self.project_config,
            task_config=task_config,
            org_config=self.org_config,
            logger=logger,
        )

    def save_exception(self, e):
        """Write an exception to disk for later analysis"""
        exception_file = self.working_dir / "exception.txt"
        exception_file.write_text(format_exc())

    def run(self):
        """The main code that runs in a sub-thread or sub-process"""
        with self.make_logger() as (logger, logfile):
            try:
                self.subtask = self._make_task(self.task_class, logger)
                self.subtask()
                logger.info(str(self.subtask.return_values))
                logger.info("SubTask Success!")
                self.results_reporter.put(
                    {
                        "status": "success",
                        "results": self.subtask.return_values,
                        "directory": str(self.working_dir),
                    }
                )
            except BaseException as e:
                logger.info(f"Failure detected: {e}")
                self.save_exception(e)
                self.failures_dir.mkdir(exist_ok=True)
                logfile.close()
                shutil.move(str(self.working_dir), str(self.failures_dir))
                self.results_reporter.put({"status": "error", "error": str(e)})
                raise

        try:
            self.outbox_dir.mkdir(exist_ok=True)
            shutil.move(str(self.working_dir), str(self.outbox_dir))
        except BaseException as e:
            self.save_exception(e)
            raise

    @contextmanager
    def make_logger(self):
        """Log to a file for potential later inspection"""
        filename = self.working_dir / f"{self.task_class.__name__}.log"
        with filename.open("w", encoding="utf-8") as f:
            logger = logging.Logger(self.task_class.__name__)

            formatter = logging.Formatter(fmt="%(asctime)s: %(message)s")
            handler = logging.StreamHandler(stream=f)
            handler.setLevel(logging.DEBUG)
            handler.setFormatter(formatter)
            logger.addHandler(handler)
            logger.propagate = False
            yield logger, f


def run_task_in_worker(worker_dict: dict, results_reporter: Queue):
    worker = TaskWorker(worker_dict, results_reporter)
    return worker.run()


def simplify(x):
    if isinstance(x, Path):
        return str(x)


class ParallelWorker:
    """Representation of the worker in the controller processs"""

    def __init__(
        self, spawn_class, worker_config: WorkerConfig, results_reporter: Queue
    ):
        self.spawn_class = spawn_class
        self.worker_config = worker_config
        self.results_reporter = results_reporter

    def _validate_worker_config_is_simple(self, worker_config):
        assert json.dumps(worker_config, default=simplify)

    def start(self):
        """Simplify config to 'json'-like datastructure, and pass to sub-process"""
        dct = self.worker_config.as_dict()
        self._validate_worker_config_is_simple(dct)

        # under the covers, Python will pass this as Pickles.
        self.process = self.spawn_class(
            target=run_task_in_worker,
            args=[dct, self.results_reporter],
            # quit if the parent process decides to exit (e.g. after a timeout)
            daemon=True,
        )
        self.process.start()

    def is_alive(self) -> bool:
        return self.process.is_alive()

    def terminate(self):
        # Note that this will throw an exception for threads
        # and should be used carefully for processes because
        # they won't necesssarily cleanup tempdirs and other
        # resources.
        self.process.terminate()

    def __repr__(self):
        return f"<Worker {self.worker_config.task_class.__name__} {self.worker_config.working_dir.name} Alive: {self.is_alive()}>"


class SubprocessKeyChain(T.NamedTuple):
    """A pretend, in-memory keychain that knows about connected apps and nothing else."""

    connected_app: T.Any = None

    def get_service(self, name):
        if name == "connected_app" and self.connected_app:
            return self.connected_app

        raise ServiceNotConfigured(name)

    def set_org(self, *args):
        pass



from collections import defaultdict
from unittest import mock

import pytest

from cumulusci.core.config.OrgConfig import OrgConfig, VersionInfo
from cumulusci.core.config.project_config import BaseProjectConfig
from cumulusci.core.config.universal_config import UniversalConfig
from cumulusci.core.exceptions import ApexTestException, SalesforceException
from cumulusci.tasks.salesforce import PackageUpload

from .util import create_task


class TestPackageUpload:
    def test_run_task(self):
        task = create_task(
            PackageUpload,
            {
                "name": "Test Release",
                "description": "Description",
                "password": "pw",
                "post_install_url": "http://www.salesforce.org",
                "release_notes_url": "https://github.com",
            },
        )

        def _init_class():
            task.tooling = mock.Mock(
                query=mock.Mock(
                    side_effect=[
                        # Query for package by namespace
                        {"totalSize": 1, "records": [{"Id": "PKG_ID"}]},
                        # Query for upload status
                        {
                            "totalSize": 1,
                            "records": [
                                {
                                    "Status": "SUCCESS",
                                    "MetadataPackageVersionId": "VERSION_ID",
                                }
                            ],
                        },
                        # Query for packge version details
                        {
                            "totalSize": 1,
                            "records": [
                                {
                                    "MajorVersion": 1,
                                    "MinorVersion": 0,
                                    "PatchVersion": 1,
                                    "ReleaseState": "Beta",
                                    "BuildNumber": 1,
                                }
                            ],
                        },
                    ]
                )
            )

        task._init_class = _init_class
        task._get_tooling_object = mock.Mock(
            return_value=mock.Mock(create=mock.Mock(return_value={"id": "UPLOAD_ID"}))
        )
        task()
        assert "SUCCESS" == task.upload["Status"]

    def test_set_package_id(self):
        name = "Test Release"
        expected_package_id = "1234567890"

        task = create_task(PackageUpload, {"name": name})
        task._get_one_record = mock.Mock(return_value={"Id": expected_package_id})

        task._set_package_id()
        assert expected_package_id == task.package_id

    def test_set_package_info(self):
        expected_package_id = "12345"
        options = {
            "name": "Test Release",
            "production": False,
            "description": "Test Description",
            "password": "secret",
            "post_install_url": "post.install.url",
            "release_notes_url": "release.notes.url",
        }

        task = create_task(PackageUpload, options)
        task._get_one_record = mock.Mock(return_value={"Id": expected_package_id})

        with pytest.raises(AttributeError):
            task.package_info

        task._set_package_info()

        assert options["name"] == task.package_info["VersionName"]
        assert options["production"] == task.package_info["IsReleaseVersion"]
        assert expected_package_id == task.package_info["MetadataPackageId"]
        assert options["description"] == task.package_info["Description"]
        assert options["password"] == task.package_info["Password"]
        assert options["post_install_url"] == task.package_info["PostInstallUrl"]
        assert options["release_notes_url"] == task.package_info["ReleaseNotesUrl"]

    @mock.patch("cumulusci.tasks.salesforce.package_upload.datetime")
    def test_make_package_upload_request(self, datetime):
        upload_start_time = "1970-01-01T00:00:00.0000Z"
        datetime.utcnow = mock.Mock(return_value=upload_start_time)

        upload_id = "asdf12345"
        package_upload_request = mock.Mock(
            create=mock.Mock(return_value={"id": upload_id})
        )

        package_id = "1234567890"
        task = create_task(PackageUpload, {"name": "Test Release"})
        task._poll = mock.Mock()
        task.package_info = {}
        task.package_id = package_id
        task.logger = mock.Mock(info=mock.Mock())
        task._get_tooling_object = mock.Mock(return_value=package_upload_request)

        task._make_package_upload_request()

        assert task._upload_start_time == upload_start_time
        assert task.logger.info.called_once_with(
            f"Created PackageUploadRequest {upload_id} for Package {package_id}"
        )

    def test_log_package_upload_errors(self):
        task = create_task(PackageUpload, {"name": "Test Release"})

        task.logger = mock.Mock(error=mock.Mock())
        task._log_apex_test_failures = mock.Mock()

        task.upload = {
            "Errors": {
                "errors": [{"message": "DmlException"}, {"message": "DmlException"}]
            }
        }
        with pytest.raises(SalesforceException):
            task._log_package_upload_errors()

        assert task.logger.error.call_count == 3
        assert not task._log_apex_test_failures.called

    def test_get_exception_type(self):
        task = create_task(PackageUpload, {"name": "Test Release"})

        apex_test_error = {"message": "ApexTestFailure"}
        assert task._get_exception_type(apex_test_error) == ApexTestException

        other_salesforce_error = {"message": "DmlException"}
        assert task._get_exception_type(other_salesforce_error) == SalesforceException

    def test_handle_apex_test_failures(self):
        task = create_task(PackageUpload, {"name": "Test Release"})
        task.logger = mock.Mock(error=mock.Mock())
        task._get_apex_test_results_from_upload = mock.Mock()
        task._log_failures = mock.Mock()

        task._handle_apex_test_failures()

        assert task.logger.error.called_once_with("Failed Apex Test")
        assert task._get_apex_test_results_from_upload.call_count == 1
        assert task._log_failures.call_count == 1

    @mock.patch("cumulusci.tasks.salesforce.package_upload.CliTable")
    def test_log_failures(self, table):
        table.echo = mock.Mock()

        task = create_task(PackageUpload, {"name": "Test Release"})

        table_data = [1, 2, 3, 4]
        task._get_table_data = mock.Mock(return_value="[1,2,3,4]")

        results = "Test Results"
        task._log_failures(results)

        assert table.called_once_with(
            table_data,
            "Failed Apex Tests",
        )
        assert table.echo.called_once()

    def test_get_table_data(self):
        task = create_task(PackageUpload, {"name": "Test Release"})

        results = {
            "records": [
                {
                    "ApexClass": {"Name": "Class1"},
                    "MethodName": "Method1",
                    "Message": "Message1",
                    "StackTrace": "StackTrace1",
                },
                {
                    "ApexClass": {"Name": "Class2"},
                    "MethodName": "Method2",
                    "Message": "Message2",
                    "StackTrace": "StackTrace2",
                },
                {
                    "ApexClass": {"Name": "Class3"},
                    "MethodName": "Method3",
                    "Message": "Message3",
                    "StackTrace": "StackTrace3",
                },
            ]
        }

        table_data = task._get_table_data(results)

        expected_table_data = [
            ["Class", "Method", "Message", "Stacktrace"],
            ["Class1", "Method1", "Message1", "StackTrace1"],
            ["Class2", "Method2", "Message2", "StackTrace2"],
            ["Class3", "Method3", "Message3", "StackTrace3"],
        ]
        assert expected_table_data == table_data

    def test_get_failed_tests_soql_qeury(self):
        task = create_task(PackageUpload, {"name": "Test Release"})
        iso_datetime = "1970-01-01T00:00:00.000"
        task._upload_start_time = mock.Mock(
            isoformat=mock.Mock(return_value=iso_datetime)
        )

        returned_query = task._get_failed_tests_soql_query()

        assert returned_query.endswith(f"TestTimestamp > {iso_datetime}Z")

    def test_run_task__upload_error(self):
        task = create_task(PackageUpload, {"name": "Test Release"})

        def _init_class():
            task._handle_apex_test_failures = mock.Mock()
            task.tooling = mock.Mock(
                query=mock.Mock(
                    side_effect=[
                        # Query for package by namespace
                        {"totalSize": 1, "records": [{"Id": "PKG_ID"}]},
                        # Query for upload status
                        {
                            "totalSize": 1,
                            "records": [
                                {
                                    "Status": "ERROR",
                                    "Errors": {
                                        "errors": [{"message": "ApexTestFailure"}]
                                    },
                                }
                            ],
                        },
                    ]
                )
            )

        task._init_class = _init_class
        task._get_tooling_object = mock.Mock(
            return_value=mock.Mock(create=mock.Mock(return_value={"id": "UPLOAD_ID"}))
        )
        with pytest.raises(ApexTestException):
            task()
        assert "ERROR" == task.upload["Status"]

    def test_get_one__no_result(self):
        task = create_task(PackageUpload, {"name": "Test Release"})
        task.tooling = mock.Mock(query=mock.Mock(return_value={"totalSize": 0}))
        with pytest.raises(SalesforceException):
            task._get_one_record(None, None)

    def test_set_return_values(self):
        task = create_task(PackageUpload, {"name": "Test Release"})

        task.package_id = "package_id"
        task.version_id = "003000000000000"
        task.version_number = "1.2.3"

        task._set_return_values()

        assert task.return_values is not None
        assert task.return_values["package_id"] == task.package_id
        assert task.return_values["version_id"] == task.version_id
        assert task.return_values["version_number"] == task.version_number

    def test_set_dependencies(self):
        project_config = BaseProjectConfig(UniversalConfig())
        project_config.project__dependencies = [{"namespace": "foo", "version": "1.0"}]
        org_config = OrgConfig({}, "test")
        org_config._installed_packages = defaultdict(list)
        org_config._installed_packages["foo@1.0"] = [
            VersionInfo("04t000000000000", "1.0")
        ]

        task = create_task(
            PackageUpload,
            {"name": "Test Release"},
            project_config=project_config,
            org_config=org_config,
        )

        task._set_dependencies()

        assert task.return_values["dependencies"] == [{"version_id": "04t000000000000"}]

    def test_log_package_upload_success(self):
        task = create_task(PackageUpload, {"name": "Test Release"})

        task.logger = mock.Mock(info=mock.Mock())

        version_id = "version_id"
        version_number = "version_number"

        task.version_id = version_id
        task.version_number = version_number

        task._log_package_upload_success()

        assert task.logger.info.called_once_with(
            f"Uploaded package version {version_number} with Id {version_id}"
        )

    def test_set_package_version_values_on_self(self):
        task = create_task(PackageUpload, {"name": "Test Release"})

        version_id = "version_id"
        major_version = 1
        minor_version = 2
        patch_version = 3
        build_number = 4

        task.upload = {"MetadataPackageVersionId": version_id}
        task._get_one_record = mock.Mock(
            return_value={
                "MajorVersion": major_version,
                "MinorVersion": minor_version,
                "PatchVersion": patch_version,
                "ReleaseState": "Beta",
                "BuildNumber": build_number,
            }
        )

        task._set_package_version_values_on_self()

        assert task.version_id == version_id
        assert (
            f"{major_version}.{minor_version}.{patch_version} (Beta {build_number})"
            == task.version_number
        )



from simple_salesforce.exceptions import SalesforceMalformedRequest

from cumulusci.core.tasks import BaseSalesforceTask
from cumulusci.core.utils import process_bool_arg
from cumulusci.tasks.salesforce.BaseSalesforceApiTask import BaseSalesforceApiTask


class CheckMyDomainActive(BaseSalesforceTask):
    def _run_task(self):
        self.return_values = (
            ".my." in self.org_config.instance_url
            or ".cloudforce.com" in self.org_config.instance_url
        )

        self.logger.info(
            f"Completed My Domain preflight check with result {self.return_values}"
        )


class CheckSettingsValue(BaseSalesforceApiTask):
    task_options = {
        "settings_type": {
            "description": "The API name of the Settings entity to be checked, such as ChatterSettings.",
            "required": True,
        },
        "settings_field": {
            "description": "The API name of the field on the Settings entity to check.",
            "required": True,
        },
        "value": {"description": "The value to check for", "required": True},
        "treat_missing_as_failure": {
            "description": "If True, treat a missing Settings entity as a preflight failure, instead of raising an exception. Defaults to False.",
            "required": False,
        },
    }

    def _run_task(self):
        field = self.options["settings_field"]
        entity = self.options["settings_type"]
        try:
            results = self.tooling.query(f"SELECT {field} FROM {entity}")["records"]
        except SalesforceMalformedRequest as e:
            self.logger.error(
                f"The settings value {entity}.{field} could not be queried: {e}"
            )
            self.return_values = False

            if not process_bool_arg(
                self.options.get("treat_missing_as_failure", False)
            ):
                raise e

            return

        if not results:
            self.logger.info(
                "Located no Settings records. Returning negative preflight result."
            )
            self.return_values = False
            return

        value = results[0].get(self.options["settings_field"])
        # Type-sensitive compare.
        if type(value) is bool:
            comparand = process_bool_arg(self.options["value"])
        elif type(value) is float:
            comparand = float(self.options["value"])
        elif type(value) is int:
            comparand = int(self.options["value"])
        else:
            comparand = self.options["value"]

        self.return_values = value == comparand

        self.logger.info(
            f"Completed Settings preflight check with result {self.return_values}"
        )



import datetime
import http.client
import io
import unittest
from collections import defaultdict
from xml.dom.minidom import parseString

import pytest
import responses
from requests import Response

from cumulusci.core.config import TaskConfig
from cumulusci.core.exceptions import ApexTestException, CumulusCIException
from cumulusci.core.tasks import BaseTask
from cumulusci.salesforce_api.exceptions import (
    MetadataApiError,
    MetadataComponentFailure,
    MetadataParseError,
)
from cumulusci.salesforce_api.metadata import (
    ApiDeploy,
    ApiListMetadata,
    ApiRetrieveInstalledPackages,
    ApiRetrievePackaged,
    ApiRetrieveUnpackaged,
    BaseMetadataApiCall,
)
from cumulusci.salesforce_api.package_zip import (
    BasePackageZipBuilder,
    CreatePackageZipBuilder,
    InstallPackageZipBuilder,
)
from cumulusci.salesforce_api.tests.metadata_test_strings import (
    deploy_result,
    deploy_result_failure,
    deploy_status_envelope,
    list_metadata_result,
    list_metadata_result_bad_val,
    list_metadata_start_envelope,
    result_envelope,
    retrieve_packaged_start_envelope,
    retrieve_result,
    retrieve_unpackaged_start_envelope,
    status_envelope,
)
from cumulusci.tests.util import DummyOrgConfig, create_project_config


class DummyPackageZipBuilder(BasePackageZipBuilder):
    def _populate_zip(self):
        return


class BaseTestMetadataApi(unittest.TestCase):
    api_class = BaseMetadataApiCall
    envelope_start = None
    envelope_status = status_envelope
    envelope_result = result_envelope

    def setUp(self):

        # Set up the mock values
        self.repo_name = "TestRepo"
        self.repo_owner = "TestOwner"
        self.repo_api_url = "https://api.github.com/repos/{}/{}".format(
            self.repo_owner, self.repo_name
        )
        self.branch = "main"

        # Create the project config
        self.project_config = create_project_config(self.repo_name, self.repo_owner)

        if not self.envelope_start:
            self.envelope_start = self.api_class.soap_envelope_start

    def _create_task(self, task_config=None, org_config=None):
        if not task_config:
            task_config = {}
        if not org_config:
            org_config = {}
        task = BaseTask(
            project_config=self.project_config,
            task_config=TaskConfig(task_config),
            org_config=DummyOrgConfig(org_config),
        )
        return task

    def _mock_call_mdapi(self, api, response, status_code=None):
        if not status_code:
            status_code = 200
        responses.add(
            method=responses.POST,
            url=api._build_endpoint_url(),
            body=response,
            status=status_code,
            content_type="text/xml; charset=utf-8",
        )
        return response

    def _create_instance(self, task, api_version=None):
        return self.api_class(task, api_version=api_version)

    def test_init(self):
        task = self._create_task()
        api = self._create_instance(task)
        self.assertEqual(api.task, task)
        self.assertEqual(
            api.api_version, self.project_config.project__package__api_version
        )

    def test_build_endpoint_url(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        self.assertEqual(
            api._build_endpoint_url(),
            "{}/services/Soap/m/{}/{}".format(
                org_config["instance_url"],
                self.project_config.project__package__api_version,
                task.org_config.org_id,
            ),
        )

    def test_build_endpoint_url_mydomain(self):
        org_config = {
            "instance_url": "https://test-org.na12.my.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        self.assertEqual(
            api._build_endpoint_url(),
            "https://test-org.na12.my.salesforce.com/services/Soap/m/{}/{}".format(
                self.project_config.project__package__api_version,
                task.org_config.org_id,
            ),
        )

    def test_build_endpoint_url_apiversion(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
        }
        task = self._create_task(org_config=org_config)
        api_version = "43.0"
        api = self._create_instance(task, api_version=api_version)
        self.assertEqual(
            api._build_endpoint_url(),
            "{}/services/Soap/m/{}/{}".format(
                org_config["instance_url"], api_version, task.org_config.org_id
            ),
        )

    def test_build_envelope_result(self):
        task = self._create_task()
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_result:
            api.soap_envelope_result = "{process_id}"
            expected = "123"
        else:
            expected = self.envelope_result.format(process_id="123")
        api.process_id = "123"
        self.assertEqual(api._build_envelope_result(), expected)

    def test_build_envelope_start(self):
        task = self._create_task()
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
            expected = str(self.project_config.project__package__api_version)
        else:
            expected = self._expected_envelope_start()

        self.assertEqual(api._build_envelope_start(), expected)

    def _expected_envelope_start(self):
        return self.envelope_start.format(
            api_version=self.project_config.project__package__api_version
        )

    def test_build_envelope_status(self):
        task = self._create_task()
        api = self._create_instance(task)
        process_id = "123"
        if not self.api_class.soap_envelope_status:
            api.soap_envelope_status = "{process_id}"
            expected = process_id
        else:
            expected = self.envelope_status.format(process_id=process_id)
        api.process_id = process_id
        self.assertEqual(api._build_envelope_status(), expected)

    def test_build_headers(self):
        action = "foo"
        message = "12345678"
        task = self._create_task()
        api = self._create_instance(task)
        self.assertEqual(
            api._build_headers(action, message),
            {
                u"Content-Type": u"text/xml; charset=UTF-8",
                u"Content-Length": "8",
                u"SOAPAction": u"foo",
            },
        )

    @responses.activate
    def test_call_faultcode(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
        response = '<?xml version="1.0" encoding="UTF-8"?><faultcode>foo</faultcode>'
        self._mock_call_mdapi(api, response)
        with self.assertRaises(MetadataApiError):
            api()

    @responses.activate
    def test_call_success(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
        if not self.api_class.soap_envelope_status:
            api.soap_envelope_status = "{process_id}"
        if not self.api_class.soap_envelope_result:
            api.soap_envelope_result = "{process_id}"

        response = '<?xml version="1.0" encoding="UTF-8"?><id>1234567890</id>'
        self._mock_call_mdapi(api, response)
        response_status = '<?xml version="1.0" encoding="UTF-8"?><done>true</done>'
        self._mock_call_mdapi(api, response_status)
        response_result = '<?xml version="1.0" encoding="UTF-8"?><foo>bar</foo>'
        response_result = self._response_call_success_result(response_result)
        self._mock_call_mdapi(api, response_result)

        resp = api()
        expected_resp = self._expected_call_success_result(response_result)

        self.assertEqual(resp, expected_resp)

    def _expected_call_success_result(self, response_result):
        return response_result

    def _response_call_success_result(self, response_result):
        return response_result

    def test_get_element_value(self):
        task = self._create_task()
        api = self._create_instance(task)
        dom = parseString("<foo>bar</foo>")
        self.assertEqual(api._get_element_value(dom, "foo"), "bar")

    def test_get_element_value_not_found(self):
        task = self._create_task()
        api = self._create_instance(task)
        dom = parseString("<foo>bar</foo>")
        self.assertEqual(api._get_element_value(dom, "baz"), None)

    def test_get_element_value_empty(self):
        task = self._create_task()
        api = self._create_instance(task)
        dom = parseString("<foo />")
        self.assertEqual(api._get_element_value(dom, "foo"), None)

    def test_get_check_interval(self):
        task = self._create_task()
        api = self._create_instance(task)
        api.check_num = 1
        self.assertEqual(api._get_check_interval(), 1)
        api.check_num = 10
        self.assertEqual(api._get_check_interval(), 4)

    @responses.activate
    def test_get_response_faultcode(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
        response = '<?xml version="1.0" encoding="UTF-8"?><faultcode>foo</faultcode>'
        self._mock_call_mdapi(api, response)
        with self.assertRaises(MetadataApiError):
            api._get_response()

    @responses.activate
    def test_get_response_faultcode_and_string(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
        response = '<?xml version="1.0" encoding="UTF-8"?>'
        response += "\n<test>"
        response += "\n  <faultcode>foo</faultcode>"
        response += "\n  <faultstring>bar</faultstring>"
        response += "\n</test>"
        self._mock_call_mdapi(api, response)
        with self.assertRaises(MetadataApiError):
            api._get_response()

    @responses.activate
    def test_get_response_faultcode_invalid_session_no_refresh(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
        response = '<?xml version="1.0" encoding="UTF-8"?><faultcode>sf:INVALID_SESSION_ID</faultcode>'
        self._mock_call_mdapi(api, response)
        with self.assertRaises(MetadataApiError):
            api._get_response()
        self.assertEqual(api.status, "Failed")

    @responses.activate
    def test_get_response_faultcode_invalid_session_refresh(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
            "refresh_token": "abcdefghij",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
        if not self.api_class.soap_envelope_status:
            api.soap_envelope_status = "{process_id}"

        mock_responses = []
        mock_responses.append(b'<?xml version="1.0" encoding="UTF-8"?><id>123</id>')
        mock_responses.append(
            b'<?xml version="1.0" encoding="UTF-8"?><faultcode>sf:INVALID_SESSION_ID</faultcode>'
        )
        mock_responses.append(b'<?xml version="1.0" encoding="UTF-8"?><foo>bar</foo>')
        for response in mock_responses:
            self._mock_call_mdapi(api, response)

        resp = api._get_response()
        self.assertEqual(resp.content, mock_responses[2])

    @responses.activate
    def test_get_response_start_error_500(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        status_code = http.client.INTERNAL_SERVER_ERROR  # HTTP Error 500
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
        if not self.api_class.soap_envelope_status:
            api.soap_envelope_status = "{process_id}"

        response = '<?xml version="1.0" encoding="UTF-8"?><foo>start</foo>'
        self._mock_call_mdapi(api, response, status_code)

        with self.assertRaises(MetadataApiError):
            api._get_response()

    @responses.activate
    def test_get_response_status_error_500(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        status_code = http.client.INTERNAL_SERVER_ERROR  # HTTP Error 500
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
        if not self.api_class.soap_envelope_status:
            api.soap_envelope_status = "{process_id}"

        response = '<?xml version="1.0" encoding="UTF-8"?><id>1234567890</id>'
        self._mock_call_mdapi(api, response)
        self._mock_call_mdapi(api, response, status_code)

        with self.assertRaises(MetadataApiError):
            api._get_response()

    @responses.activate
    def test_get_response_status_no_loop(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
        if not self.api_class.soap_envelope_status:
            api.soap_envelope_status = "{process_id}"
        if not self.api_class.soap_envelope_result:
            api.soap_envelope_result = "{process_id}"

        response = b'<?xml version="1.0" encoding="UTF-8"?><id>1234567890</id>'
        self._mock_call_mdapi(api, response)
        response_status = b'<?xml version="1.0" encoding="UTF-8"?><done>true</done>'
        self._mock_call_mdapi(api, response_status)
        response_result = b'<?xml version="1.0" encoding="UTF-8"?><foo>bar</foo>'
        self._mock_call_mdapi(api, response_result)

        resp = api._get_response()

        self.assertEqual(resp.content, response_result)

    @responses.activate
    def test_get_response_status_loop_twice(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
        if not self.api_class.soap_envelope_status:
            api.soap_envelope_status = "{process_id}"
        if not self.api_class.soap_envelope_result:
            api.soap_envelope_result = "{process_id}"
        api.check_interval = 0

        response = b'<?xml version="1.0" encoding="UTF-8"?><id>1234567890</id>'
        self._mock_call_mdapi(api, response)
        response_status = b'<?xml version="1.0" encoding="UTF-8"?><done>false</done>'
        self._mock_call_mdapi(api, response_status)
        response_status = b'<?xml version="1.0" encoding="UTF-8"?><done>false</done>'
        self._mock_call_mdapi(api, response_status)
        response_status = b'<?xml version="1.0" encoding="UTF-8"?><done>true</done>'
        self._mock_call_mdapi(api, response_status)
        response_result = b'<?xml version="1.0" encoding="UTF-8"?><foo>bar</foo>'
        self._mock_call_mdapi(api, response_result)

        resp = api._get_response()

        self.assertEqual(resp.content, response_result)

        self.assertEqual(api.status, "Done")

        self.assertEqual(api.check_num, 4)

    def test_process_response_status_no_done_element(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            b'<?xml version="1.0" encoding="UTF-8"?><foo>status</foo>'
        )
        res = api._process_response_status(response)
        self.assertEqual(api.status, "Failed")
        self.assertEqual(res.content, response.content)

    def test_process_response_status_done_is_true(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            b'<?xml version="1.0" encoding="UTF-8"?><done>true</done>'
        )
        res = api._process_response_status(response)
        self.assertEqual(api.status, "Done")
        self.assertEqual(res.content, response.content)

    def test_process_response_status_pending(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            b'<?xml version="1.0" encoding="UTF-8"?><done>false</done>'
        )
        res = api._process_response_status(response)
        self.assertEqual(api.status, "Pending")
        self.assertEqual(res.content, response.content)

    def test_process_response_status_in_progress(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            b'<?xml version="1.0" encoding="UTF-8"?><done>false</done>'
        )
        api.status = "InProgress"
        res = api._process_response_status(response)
        self.assertEqual(api.status, "InProgress")
        self.assertEqual(res.content, response.content)

    def test_process_response_status_in_progress_state_detail(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            b'<?xml version="1.0" encoding="UTF-8"?><test><done>false</done><stateDetail>Deploy log goes here</stateDetail></test>'
        )
        api.status = "InProgress"
        res = api._process_response_status(response)
        self.assertEqual(api.status, "InProgress")
        self.assertEqual(res.content, response.content)


class TestBaseMetadataApiCall(BaseTestMetadataApi):
    def test_build_envelope_start_no_envelope(self):
        task = self._create_task()
        api = self._create_instance(task)
        with pytest.raises(AssertionError):
            api._build_envelope_start()

    def test_build_envelope_status_no_envelope(self):
        task = self._create_task()
        api = self._create_instance(task)
        self.assertEqual(api._build_envelope_status(), None)

    def test_build_envelope_result_no_envelope(self):
        task = self._create_task()
        api = self._create_instance(task)
        self.assertEqual(api._build_envelope_result(), None)

    def test_get_response_no_start_env(self):
        task = self._create_task()
        api = self._create_instance(task)
        with self.assertRaises(NotImplementedError):
            api._get_response()

    @responses.activate
    def test_get_response_no_status(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        api.soap_envelope_start = "{api_version}"
        response = b'<?xml version="1.0" encoding="UTF-8"?><foo />'
        self._mock_call_mdapi(api, response)
        resp = api._get_response()
        self.assertEqual(resp.content, response)


class TestApiDeploy(BaseTestMetadataApi):
    api_class = ApiDeploy
    envelope_status = deploy_status_envelope

    def setUp(self):
        super(TestApiDeploy, self).setUp()
        self.package_zip = DummyPackageZipBuilder().as_base64()

    def _expected_envelope_start(self):
        return self.envelope_start.format(
            package_zip=self.package_zip,
            check_only="false",
            purge_on_delete="false",
            test_level="",
            run_tests="",
        )

    def _response_call_success_result(self, response_result):
        return deploy_result.format(status="Succeeded", extra="").encode()

    def _expected_call_success_result(self, response_result):
        return "Success"

    def _create_instance(
        self,
        task,
        api_version=None,
        purge_on_delete=None,
        check_only=None,
        test_level=None,
        run_tests=None,
    ):
        return self.api_class(
            task,
            self.package_zip,
            api_version=api_version,
            purge_on_delete=purge_on_delete,
            check_only=check_only,
            test_level=test_level,
            run_tests=run_tests,
        )

    def test_init_no_purge_on_delete(self):
        task = self._create_task()
        api = self._create_instance(task, purge_on_delete=False)
        self.assertEqual(api.purge_on_delete, "false")

    def test_init_default_check_only(self):
        task = self._create_task()
        api = self._create_instance(task)
        self.assertEqual(api.check_only, "false")

    def test_init_check_only(self):
        task = self._create_task()
        api = self._create_instance(task, check_only=True)
        self.assertEqual(api.check_only, "true")

    def test_init_default_test_level(self):
        task = self._create_task()
        api = self._create_instance(task)
        self.assertEqual(api.test_level, None)

    def test_init_test_level(self):
        task = self._create_task()
        api = self._create_instance(task, test_level="NoTestRun")
        self.assertEqual(api.test_level, "NoTestRun")

    def test_init_default_run_tests(self):
        task = self._create_task()
        api = self._create_instance(task)
        self.assertEqual(api.run_tests, [])

    def test_init_run_tests(self):
        task = self._create_task()
        api = self._create_instance(task, run_tests=["TestA", "TestB"])
        self.assertEqual(api.run_tests, ["TestA", "TestB"])

    def test_build_envelope_status__run_specified_tests(self):
        task = self._create_task()
        api = self._create_instance(
            task, run_tests=["TestA", "TestB"], test_level="RunSpecifiedTests"
        )

        api.package_zip = "Test"

        envelope = api._build_envelope_start()
        assert "<runTests>TestA</runTests>" in envelope
        assert "<runTests>TestB</runTests>" in envelope
        assert "RunSpecifiedTests" in envelope

    def test_process_response_metadata_failure(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            deploy_result_failure.format(
                details="""<componentFailures>
  <problem>problem</problem>
  <problemType>Error</problemType>
  <componentType>CustomObject</componentType>
  <fileName>Test__c</fileName>
  <lineNumber>1</lineNumber>
  <columnNumber>1</columnNumber>
  <created>false</created>
  <deleted>false</deleted>
</componentFailures>"""
            ).encode()
        )
        with self.assertRaises(MetadataComponentFailure) as cm:
            api._process_response(response)
        expected = "Update of CustomObject Test__c: Error on line 1, col 1: problem"
        self.assertEqual(expected, str(cm.exception))

    def test_process_response_metadata_failure_no_lineno(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            deploy_result_failure.format(
                details="""<componentFailures>
  <problem>problem</problem>
  <problemType>Error</problemType>
  <componentType>CustomObject</componentType>
  <fileName>Test__c</fileName>
  <created>false</created>
  <deleted>false</deleted>
</componentFailures>"""
            ).encode()
        )
        with self.assertRaises(MetadataComponentFailure) as cm:
            api._process_response(response)
        expected = "Update of CustomObject Test__c: Error: problem"
        self.assertEqual(expected, str(cm.exception))

    def test_process_response_metadata_failure_no_file_name(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            deploy_result_failure.format(
                details="""<componentFailures>
  <problem>problem</problem>
  <problemType>Error</problemType>
  <componentType>CustomObject</componentType>
  <created>false</created>
  <deleted>false</deleted>
</componentFailures>"""
            ).encode()
        )
        with self.assertRaises(MetadataComponentFailure) as cm:
            api._process_response(response)
        expected = "Update of CustomObject: Error: problem"
        self.assertEqual(expected, str(cm.exception))

    def test_process_response_problem(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            deploy_result_failure.format(
                details="""<problem>problem</problem>"""
            ).encode()
        )
        with self.assertRaises(MetadataApiError) as cm:
            api._process_response(response)
        expected = "problem"
        self.assertEqual(expected, str(cm.exception))

    def test_process_response_test_failure(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            deploy_result_failure.format(
                details="""<runTestResult>
  <failures>
    <namespace>test</namespace>
    <stackTrace>stack</stackTrace>
  </failures>
</runTestResult>
"""
            ).encode()
        )
        with self.assertRaises(ApexTestException) as cm:
            api._process_response(response)
        expected = "Apex Test Failure: from namespace test: stack"
        self.assertEqual(expected, str(cm.exception))

    def test_process_response_no_status(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(b"<bogus />")
        status = api._process_response(response)
        self.assertEqual(status, "Failed")

    def test_process_response_failure_but_no_message(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(b"<status>Failed</status>")
        with self.assertRaises(MetadataApiError) as cm:
            api._process_response(response)
        self.assertEqual(response.text, str(cm.exception))

    def test_get_action(self):
        task = self._create_task()
        api = self._create_instance(task)
        self.assertEqual("Create", api._get_action(True, False))
        self.assertEqual("Delete", api._get_action(False, True))
        self.assertEqual("Update", api._get_action(False, False))


class TestApiListMetadata(BaseTestMetadataApi):
    api_class = ApiListMetadata
    envelope_start = list_metadata_start_envelope

    def setUp(self):
        super(TestApiListMetadata, self).setUp()
        self.metadata_type = "CustomObject"
        self.metadata = None
        self.folder = None
        self.api_version = self.project_config.project__package__api_version

    def _response_call_success_result(self, response_result):
        return list_metadata_result

    def _expected_call_success_result(self, response_result):
        metadata = defaultdict(list)
        metadata["CustomObject"] = [
            {
                u"createdById": None,
                u"createdByName": None,
                u"createdDate": datetime.datetime(2018, 8, 7, 16, 31, 57),
                u"fileName": None,
                u"fullName": u"Test__c",
                u"id": None,
                u"lastModifiedById": None,
                u"lastModifiedByName": None,
                u"lastModifiedDate": None,
                u"manageableState": None,
                u"namespacePrefix": None,
                u"type": "CustomObject",
            }
        ]
        return metadata

    def _create_instance(self, task, api_version=None):
        if api_version is None:
            api_version = self.api_version
        return self.api_class(
            task,
            metadata_type=self.metadata_type,
            metadata=self.metadata,
            folder=self.folder,
            as_of_version=api_version,
        )

    @responses.activate
    def test_bad_date_somehow(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)

        self._mock_call_mdapi(api, list_metadata_result_bad_val)
        with self.assertRaises(MetadataParseError):
            api()


class TestApiRetrieveUnpackaged(BaseTestMetadataApi):
    maxDiff = None
    api_class = ApiRetrieveUnpackaged
    envelope_start = retrieve_unpackaged_start_envelope

    def setUp(self):
        super(TestApiRetrieveUnpackaged, self).setUp()
        self.package_xml = """<?xml version="1.0" encoding="UTF-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
    <version>41.0</version>
</Package>"""
        self.result_zip = DummyPackageZipBuilder()

    def _response_call_success_result(self, response_result):
        return retrieve_result.format(
            zip=self.result_zip.as_base64(), extra=""
        ).encode()

    def _expected_call_success_result(self, response_result):
        return self.result_zip.zf

    def _create_instance(self, task, api_version=None):
        return self.api_class(task, self.package_xml, api_version=api_version)

    @responses.activate
    def test_call_success(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
        if not self.api_class.soap_envelope_status:
            api.soap_envelope_status = "{process_id}"
        if not self.api_class.soap_envelope_result:
            api.soap_envelope_result = "{process_id}"

        response = '<?xml version="1.0" encoding="UTF-8"?><id>1234567890</id>'
        self._mock_call_mdapi(api, response)
        response_status = '<?xml version="1.0" encoding="UTF-8"?><done>true</done>'
        self._mock_call_mdapi(api, response_status)
        response_result = '<?xml version="1.0" encoding="UTF-8"?><foo>bar</foo>'
        response_result = self._response_call_success_result(response_result)
        self._mock_call_mdapi(api, response_result)

        zip_file = api()

        self.assertEqual(
            zip_file.namelist(),
            self._expected_call_success_result(response_result).namelist(),
        )


class TestApiRetrieveInstalledPackages(BaseTestMetadataApi):
    api_class = ApiRetrieveInstalledPackages

    def _create_instance(self, task, api_version=None):
        api = self.api_class(task, api_version)
        return api

    def _expected_call_success_result(self, result_response):
        return {}

    def test_process_response_no_zipstr(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            deploy_result.format(status="testing", extra="").encode()
        )
        resp = api._process_response(response)
        self.assertEqual(resp, {})

    def test_process_response_zipstr_no_packages(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            retrieve_result.format(
                zip=CreatePackageZipBuilder("testing", api.api_version).as_base64(),
                extra="",
            ).encode()
        )
        resp = api._process_response(response)
        self.assertEqual(resp, {})

    def test_process_response_zipstr_one_package(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            retrieve_result.format(
                zip=InstallPackageZipBuilder("foo", "1.1").as_base64(), extra=""
            ).encode()
        )
        resp = api._process_response(response)
        self.assertEqual(resp, {"foo": "1.1"})


class TestApiRetrievePackaged(TestApiRetrieveUnpackaged):
    api_class = ApiRetrievePackaged
    envelope_start = retrieve_packaged_start_envelope

    def setUp(self):
        super(TestApiRetrievePackaged, self).setUp()
        self.package_name = "Test Package"

    def _expected_envelope_start(self):
        return self.envelope_start.format(
            api_version=self.project_config.project__package__api_version,
            package_name=self.package_name,
        )

    def _create_instance(self, task, api_version=None):
        return self.api_class(task, self.package_name, api_version)

    def test_process_response__no_package_match_found(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.raw = io.BytesIO(
            b"INVALID_CROSS_REFERENCE_KEY: No package named Test Package"
        )
        with self.assertRaises(CumulusCIException):
            api._process_response(response)



import json
import os
import re
from pathlib import Path

import requests

from cumulusci.core.config import BaseProjectConfig, FlowConfig, TaskConfig
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.flowrunner import FlowCoordinator
from cumulusci.core.github import get_tag_by_name
from cumulusci.core.tasks import BaseTask
from cumulusci.core.utils import process_bool_arg
from cumulusci.utils import cd, download_extract_github, temporary_dir
from cumulusci.utils.http.requests_utils import safe_json_from_response

INSTALL_VERSION_RE = re.compile(r"^Install .*\d$")


class BaseMetaDeployTask(BaseTask):
    """Base class for tasks that talk to MetaDeploy's API."""

    def _init_task(self):
        metadeploy_service = self.project_config.keychain.get_service("metadeploy")
        self.base_url = metadeploy_service.url
        self.api = requests.Session()
        self.api.headers["Authorization"] = "token {}".format(metadeploy_service.token)

    def _call_api(self, method, path, collect_pages=False, **kwargs):
        next_url = self.base_url + path
        results = []
        while next_url is not None:
            response = self.api.request(method, next_url, **kwargs)
            if response.status_code == 400:
                raise requests.exceptions.HTTPError(response.content)
            response = safe_json_from_response(response)
            if "links" in response and collect_pages:
                results.extend(response["data"])
                next_url = response["links"]["next"]
            else:
                return response
        return results


class Publish(BaseMetaDeployTask):
    """Publishes installation plans to MetaDeploy."""

    task_options = {
        "tag": {"description": "Name of the git tag to publish"},
        "commit": {"description": "Commit hash to publish"},
        "plan": {
            "description": "Name of the plan(s) to publish. "
            "This refers to the `plans` section of cumulusci.yml. "
            "By default, all plans will be published.",
            "required": False,
        },
        "dry_run": {
            "description": "If True, print steps without publishing.",
            "required": False,
        },
        "publish": {
            "description": "If True, set is_listed to True on the version. Default: False",
            "required": False,
        },
        "labels_path": {
            "description": "Path to a folder containing translations.",
            "required": False,
        },
    }

    def _init_task(self):
        super(Publish, self)._init_task()
        self.dry_run = process_bool_arg(self.options.get("dry_run") or False)
        self.publish = not self.dry_run and process_bool_arg(
            self.options.get("publish") or False
        )
        self.tag = self.options.get("tag")
        self.commit = self.options.get("commit")
        if not self.tag and not self.commit:
            raise TaskOptionsError("You must specify either the tag or commit option.")
        self.labels_path = self.options.get("labels_path", "metadeploy")
        if not os.path.exists(self.labels_path):  # pragma: no cover
            os.makedirs(self.labels_path)

        plan_name = self.options.get("plan")
        if plan_name:
            plan_configs = {}
            plan_configs[plan_name] = getattr(
                self.project_config, "plans__{}".format(plan_name)
            )
            self.plan_configs = plan_configs
        else:
            self.plan_configs = self.project_config.plans

        self._load_labels()

    def _run_task(self):
        # Find or create Version
        product = self._find_product()
        if not self.dry_run:
            version = self._find_or_create_version(product)
            if self.labels_path and "slug" in product:
                self._publish_labels(product["slug"])

        # Check out the specified tag
        repo_owner = self.project_config.repo_owner
        repo_name = self.project_config.repo_name
        repo_url = f"https://github.com/{repo_owner}/{repo_name}"
        gh = self.project_config.get_github_api()
        repo = gh.repository(repo_owner, repo_name)
        if self.tag:
            tag = get_tag_by_name(repo, self.tag)
            self.commit = tag.object.sha
        self.logger.info(
            f"Downloading commit {self.commit} of {repo.full_name} from GitHub"
        )
        zf = download_extract_github(gh, repo_owner, repo_name, ref=self.commit)
        with temporary_dir() as project_dir:
            zf.extractall(project_dir)
            project_config = BaseProjectConfig(
                self.project_config.universal_config_obj,
                repo_info={
                    "root": project_dir,
                    "owner": repo_owner,
                    "name": repo_name,
                    "url": repo_url,
                    "branch": self.tag or self.commit,
                    "commit": self.commit,
                },
            )
            project_config.set_keychain(self.project_config.keychain)

            # Create each plan
            for plan_name, plan_config in self.plan_configs.items():

                self._add_labels(
                    plan_config,
                    f"plan:{plan_name}",
                    {
                        "title": "title of installation plan",
                        "preflight_message": "shown before user starts installation (markdown)",
                        "preflight_message_additional": "shown before user starts installation (markdown)",
                        "post_install_message": "shown after successful installation (markdown)",
                        "post_install_message_additional": "shown after successful installation (markdown)",
                        "error_message": "shown after failed installation (markdown)",
                    },
                )
                checks = plan_config.get("checks") or []
                for check in checks:
                    self._add_label(
                        "checks", check.get("message"), "shown if validation fails"
                    )

                steps = self._freeze_steps(project_config, plan_config)
                self.logger.debug("Prepared steps:\n" + json.dumps(steps, indent=4))
                for step in steps:
                    # avoid separate labels for installing each package
                    if INSTALL_VERSION_RE.match(step["name"]):
                        self._add_label(
                            "steps",
                            "Install {product} {version}",
                            "title of installation step",
                        )
                    else:
                        self._add_label(
                            "steps", step["name"], "title of installation step"
                        )
                    self._add_label(
                        "steps",
                        step.get("description"),
                        "description of installation step",
                    )
                    for check in step["task_config"].get("checks", []):
                        self._add_label(
                            "checks", check.get("message"), "shown if validation fails"
                        )

                if not self.dry_run:
                    self._publish_plan(product, version, plan_name, plan_config, steps)

            # Update version to set is_listed=True
            if self.publish:
                self._call_api(
                    "PATCH",
                    "/versions/{}".format(version["id"]),
                    json={"is_listed": True},
                )
                self.logger.info("Published Version {}".format(version["url"]))

        # Save labels
        self._save_labels()

    def _publish_plan(self, product, version, plan_name, plan_config, steps):
        plan_template = self._find_or_create_plan_template(
            product, plan_name, plan_config
        )

        plan_json = {
            "is_listed": plan_config.get("is_listed", True),
            "plan_template": plan_template["url"],
            "post_install_message_additional": plan_config.get(
                "post_install_message_additional", ""
            ),
            "preflight_message_additional": plan_config.get(
                "preflight_message_additional", ""
            ),
            "steps": steps,
            "tier": plan_config["tier"],
            "title": plan_config["title"],
            "version": version["url"],
            # Use same AllowedList as the product, if any
            "visible_to": product.get("visible_to"),
        }
        if plan_config.get("checks"):
            plan_json["preflight_checks"] = plan_config["checks"]

        # Create Plan
        plan = self._call_api("POST", "/plans", json=plan_json)
        self.logger.info("Created Plan {}".format(plan["url"]))

    def _freeze_steps(self, project_config, plan_config):
        steps = plan_config["steps"]
        flow_config = FlowConfig(plan_config)
        flow_config.project_config = project_config
        flow = FlowCoordinator(project_config, flow_config)
        steps = []
        for step in flow.steps:
            if step.skip:
                continue
            with cd(step.project_config.repo_root):
                task = step.task_class(
                    step.project_config,
                    TaskConfig(step.task_config),
                    name=step.task_name,
                )
                steps.extend(task.freeze(step))
        return steps

    def _find_product(self):
        repo_url = self.project_config.project__git__repo_url
        result = self._call_api("GET", "/products", params={"repo_url": repo_url})
        if len(result["data"]) != 1:
            raise Exception(
                "No product found in MetaDeploy with repo URL {}".format(repo_url)
            )
        product = result["data"][0]
        self._add_labels(
            product,
            "product",
            {
                "title": "name of product",
                "short_description": "tagline of product",
                "description": "shown on product detail page (markdown)",
                "click_through_agreement": "legal text shown in modal dialog",
                "error_message": "shown after failed installation (markdown)",
            },
        )
        return product

    def _find_or_create_version(self, product):
        """Create a Version in MetaDeploy if it doesn't already exist"""
        if self.tag:
            label = self.project_config.get_version_for_tag(self.tag)
        else:
            label = self.commit
        result = self._call_api(
            "GET", "/versions", params={"product": product["id"], "label": label}
        )
        if len(result["data"]) == 0:
            version = self._call_api(
                "POST",
                "/versions",
                json={
                    "product": product["url"],
                    "label": label,
                    "description": self.options.get("description", ""),
                    "is_production": True,
                    "commit_ish": self.tag or self.commit,
                    "is_listed": False,
                },
            )
            self.logger.info("Created {}".format(version["url"]))
        else:
            version = result["data"][0]
            self.logger.info("Found {}".format(version["url"]))
        return version

    def _find_or_create_plan_template(self, product, plan_name, plan_config):
        result = self._call_api(
            "GET",
            "/plantemplates",
            params={"product": product["id"], "name": plan_name},
        )
        if len(result["data"]) == 0:
            plantemplate = self._call_api(
                "POST",
                "/plantemplates",
                json={
                    "name": plan_name,
                    "product": product["url"],
                    "preflight_message": plan_config.get("preflight_message", ""),
                    "post_install_message": plan_config.get("post_install_message", ""),
                    "error_message": plan_config.get("error_message", ""),
                },
            )
            self.logger.info("Created {}".format(plantemplate["url"]))
            planslug = self._call_api(
                "POST",
                "/planslug",
                json={"slug": plan_config["slug"], "parent": plantemplate["url"]},
            )
            self.logger.info("Created {}".format(planslug["url"]))
        else:
            plantemplate = result["data"][0]
            self.logger.info("Found {}".format(plantemplate["url"]))
        return plantemplate

    def _load_labels(self):
        """Load existing English labels."""
        self.labels = {}
        labels_path = os.path.join(self.labels_path, "labels_en.json")
        if os.path.exists(labels_path):
            with open(labels_path, "r") as f:
                self.labels = json.load(f)

    def _add_labels(self, obj, category, fields):
        """Add specified fields from obj to a label category."""
        for name, description in fields.items():
            text = obj.get(name)
            if text:
                if category not in self.labels:
                    self.labels[category] = {}
                label = {"message": text, "description": description}
                self.labels[category][name] = label

    def _add_label(self, category, text, description):
        """Add a single label to a label category."""
        if not text:
            return
        if category not in self.labels:
            self.labels[category] = {}
        label = {"message": text, "description": description}
        self.labels[category][text] = label

    def _save_labels(self):
        """Save updates to English labels."""
        if self.labels_path:
            labels_path = os.path.join(self.labels_path, "labels_en.json")
            self.logger.info(f"Updating labels in {labels_path}")
            with open(labels_path, "w") as f:
                json.dump(self.labels, f, indent=4)

    def _publish_labels(self, slug):
        """Publish labels in all languages to MetaDeploy."""
        for path in Path(self.labels_path).glob("*.json"):
            lang = path.stem.split("_")[-1].lower()
            if lang in ("en", "en-us"):
                continue
            orig_labels = json.loads(path.read_text())
            prefixed_labels = {}
            for context, labels in orig_labels.items():
                prefixed_labels[f"{slug}:{context}"] = labels
            self.logger.info(f"Updating {lang} translations")
            try:
                self._call_api("PATCH", f"/translations/{lang}", json=prefixed_labels)
            except requests.exceptions.HTTPError as err:
                self.logger.warning(f"Could not update {lang} translation: {err}")



import unittest

import responses

from cumulusci.core.config import BaseProjectConfig, TaskConfig, UniversalConfig
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.keychain import BaseProjectKeychain
from cumulusci.core.tests.utils import MockLoggerMixin
from cumulusci.tasks.salesforce.custom_settings_wait import CustomSettingValueWait
from cumulusci.tests.util import DummyOrgConfig


class TestRunCustomSettingsWait(MockLoggerMixin, unittest.TestCase):
    def setUp(self):
        self.api_version = 42.0
        self.universal_config = UniversalConfig(
            {"project": {"api_version": self.api_version}}
        )
        self.task_config = TaskConfig()
        self.project_config = BaseProjectConfig(
            self.universal_config, config={"noyaml": True}
        )
        self.project_config.config["project"] = {
            "package": {"api_version": self.api_version}
        }
        keychain = BaseProjectKeychain(self.project_config, "")
        self.project_config.set_keychain(keychain)
        self.org_config = DummyOrgConfig(
            {
                "id": "foo/1",
                "instance_url": "https://example.com",
                "access_token": "abc123",
            },
            "test",
        )
        self.base_normal_url = "{}/services/data/v{}/query/".format(
            self.org_config.instance_url, self.api_version
        )
        self.task_log = self._task_log_handler.messages

    def _get_query_resp(self):
        return {
            "size": 1,
            "totalSize": 1,
            "done": True,
            "queryLocator": None,
            "entityTypeName": "Customizable_Rollup_Setings__c",
            "records": [
                {
                    "attributes": {
                        "type": "Customizable_Rollup_Setings__c",
                        "url": "/services/data/v47.0/sobjects/Customizable_Rollup_Setings__c/707L0000014nnPHIAY",
                    },
                    "Id": "707L0000014nnPHIAY",
                    "SetupOwnerId": "00Dxxxxxxxxxxxx",
                    "Customizable_Rollups_Enabled__c": True,
                    "Rollups_Account_Batch_Size__c": 200,
                }
            ],
        }

    def _get_url_and_task(self):
        task = CustomSettingValueWait(
            self.project_config, self.task_config, self.org_config
        )
        url = self.base_normal_url
        return task, url

    @responses.activate
    def test_run_custom_settings_wait_match_bool(self):
        self.task_config.config["options"] = {
            "object": "Customizable_Rollup_Setings__c",
            "field": "Customizable_Rollups_Enabled__c",
            "value": True,
            "poll_interval": 1,
        }

        # simulate finding no settings record and then finding one with the expected value
        task, url = self._get_url_and_task()
        responses.add(
            responses.GET, url, json={"totalSize": 0, "done": True, "records": []}
        )
        response = self._get_query_resp()
        response["records"][0]["Customizable_Rollups_Enabled__c"] = True
        responses.add(responses.GET, url, json=response)
        task()

    @responses.activate
    def test_run_custom_settings_wait_match_bool_changed_case(self):
        self.task_config.config["options"] = {
            "object": "CUSTOMIZABLE_Rollup_Setings__c",
            "field": "CUSTOMIZABLE_Rollups_enabled__C",
            "value": True,
            "poll_interval": 1,
        }

        task, url = self._get_url_and_task()
        response = self._get_query_resp()
        response["records"][0]["Customizable_Rollups_Enabled__c"] = True
        responses.add(responses.GET, url, json=response)
        task()

    @responses.activate
    def test_run_custom_settings_wait_match_int(self):
        self.task_config.config["options"] = {
            "object": "Customizable_Rollup_Setings__c",
            "field": "Rollups_Account_Batch_Size__c",
            "value": 200,
            "poll_interval": 1,
        }

        task, url = self._get_url_and_task()
        response = self._get_query_resp()
        response["records"][0]["Rollups_Account_Batch_Size__c"] = 200
        responses.add(responses.GET, url, json=response)
        task()

    @responses.activate
    def test_run_custom_settings_wait_match_str(self):
        self.task_config.config["options"] = {
            "object": "Customizable_Rollup_Setings__c",
            "field": "Rollups_Account_Batch_Size__c",
            "value": "asdf",
            "poll_interval": 1,
        }

        task, url = self._get_url_and_task()
        response = self._get_query_resp()
        response["records"][0]["Rollups_Account_Batch_Size__c"] = "asdf"
        responses.add(responses.GET, url, json=response)
        task()

    @responses.activate
    def test_run_custom_settings_wait_not_settings_object(self):
        self.task_config.config["options"] = {
            "object": "Customizable_Rollup_Setings__c",
            "field": "Rollups_Account_Batch_Size__c",
            "value": 200,
            "poll_interval": 1,
        }

        task, url = self._get_url_and_task()
        responses.add(
            responses.GET,
            url,
            status=400,
            json=[
                {
                    "message": "\nSELECT SetupOwnerId FROM npe5__Affiliation__c\n       ^\nERROR at Row:1:Column:8\nNo such column 'SetupOwnerId' on entity 'npe5__Affiliation__c'. If you are attempting to use a custom field, be sure to append the '__c' after the custom field name. Please reference your WSDL or the describe call for the appropriate names.",
                    "errorCode": "INVALID_FIELD",
                }
            ],
        )

        with self.assertRaises(TaskOptionsError) as e:
            task()

        assert "supported" in str(e.exception)

    def test_apply_namespace__managed(self):
        self.project_config.config["project"]["package"]["namespace"] = "ns"
        self.task_config.config["options"] = {
            "object": "%%%NAMESPACE%%%Test__c",
            "field": "Field__c",
            "value": "x",
            "managed": True,
            "namespaced": True,
        }
        task, url = self._get_url_and_task()
        task.object_name = "%%%NAMESPACE%%%Test__c"
        task.field_name = "%%%NAMESPACE%%%Field__c"
        task._apply_namespace()
        assert task.object_name == "ns__Test__c"
        assert task.field_name == "ns__Field__c"



import http.client
import ssl
import sys
import threading
import time
import urllib.error
import urllib.parse
import urllib.request
from contextlib import contextmanager
from unittest import mock

import pytest
import responses
from requests.models import Response

from cumulusci.core.exceptions import SalesforceCredentialsException
from cumulusci.core.keychain.base_project_keychain import DEFAULT_CONNECTED_APP_PORT
from cumulusci.oauth.client import (
    PORT_IN_USE_ERR,
    OAuth2Client,
    OAuth2ClientConfig,
    OAuth2DeviceConfig,
    get_device_code,
    get_device_oauth_token,
)
from cumulusci.oauth.exceptions import OAuth2Error
from cumulusci.oauth.salesforce import jwt_session


@responses.activate
@mock.patch("cumulusci.oauth.salesforce.jwt.encode")
def test_jwt_session(encode):
    # Mock the call to encode so we don't need
    # to generate a private key that would be committed
    error = "Yeti"
    responses.add(
        responses.POST,
        "https://login.salesforce.com/services/oauth2/token",
        body=error,
        status=400,
    )
    with pytest.raises(
        SalesforceCredentialsException, match=f"Error retrieving access token: {error}"
    ):
        jwt_session("client_id", "server_key", "username")


@pytest.fixture
def client_config():
    return {
        "client_id": "foo_id",
        "client_secret": "foo_secret",
        "auth_uri": "https://login.salesforce.com/services/oauth2/authorize",
        "token_uri": "https://login.salesforce.com/services/oauth2/token",
        "redirect_uri": "http://localhost:7788/callback",
        "scope": "web full refresh_token",
        "prompt": "login",
    }


@pytest.fixture
def client(client_config):
    return OAuth2Client(client_config)


@pytest.fixture
def http_client(client_config):
    client_config = client_config.copy()
    client_config["redirect_uri"] = "http://localhost:8080/callback"
    return OAuth2Client(client_config)


@contextmanager
@mock.patch("time.sleep", time.sleep)  # undo mock from conftest
def httpd_thread(oauth_client):
    # call OAuth object on another thread - this spawns local httpd
    thread = threading.Thread(target=oauth_client.auth_code_flow)
    thread.start()
    while thread.is_alive():
        if oauth_client.httpd:
            break
        time.sleep(0.01)

    assert (
        oauth_client.httpd
    ), "HTTPD did not start. Perhaps port 8080 cannot be accessed."

    try:
        yield oauth_client
    finally:
        oauth_client.httpd.shutdown()
        thread.join()


@mock.patch("webbrowser.open", mock.MagicMock(return_value=None))
class TestOAuth2Client:
    @responses.activate
    def test_refresh_token(self, client):
        responses.add(
            responses.POST,
            "https://login.salesforce.com/services/oauth2/token",
            body=b'{"message":"SENTINEL"}',
        )
        info = client.refresh_token("token")
        assert "SENTINEL" == info["message"]

    @responses.activate
    def test_auth_code_flow___http(self, http_client):
        expected_response = {
            "access_token": "abc123",
            "id_token": "abc123",
            "token_type": "Bearer",
            "signature": "abc123",
            "issued_at": "12345",
            "scope": "web full refresh_token",
            "instance_url": "https://na15.salesforce.com",
            "id": "https://login.salesforce.com/id/abc/xyz",
            "refresh_token": "abc123",
        }
        responses.add(
            responses.POST,
            "https://login.salesforce.com/services/oauth2/token",
            status=http.client.OK,
            json=expected_response,
        )

        # call OAuth object on another thread - this spawns local httpd
        with httpd_thread(http_client) as oauth_client:
            # simulate callback from browser
            response = urllib.request.urlopen(
                http_client.client_config.redirect_uri + "?code=123"
            )

        assert oauth_client.response.json() == expected_response
        assert b"Congratulations" in response.read()

    @responses.activate
    def test_auth_code_flow___https(self, client):
        expected_response = {
            "access_token": "abc123",
            "id_token": "abc123",
            "token_type": "Bearer",
            "signature": "abc123",
            "issued_at": "12345",
            "scope": "web full refresh_token",
            "instance_url": "https://na15.salesforce.com",
            "id": "https://login.salesforce.com/id/abc/xyz",
            "refresh_token": "abc123",
        }
        responses.add(
            responses.POST,
            "https://login.salesforce.com/services/oauth2/token",
            status=http.client.OK,
            json=expected_response,
        )
        # use https for callback
        client.client_config.redirect_uri = "https://localhost:8080/callback"
        # squash CERTIFICATE_VERIFY_FAILED from urllib
        # https://stackoverflow.com/questions/49183801/ssl-certificate-verify-failed-with-urllib
        ssl._create_default_https_context = ssl._create_unverified_context

        # call OAuth object on another thread - this spawns local httpd
        with httpd_thread(client) as oauth_client:
            # simulate callback from browser
            response = urllib.request.urlopen(
                oauth_client.client_config.redirect_uri + "?code=123"
            )

        assert oauth_client.response.json() == expected_response
        assert b"Congratulations" in response.read()

    @responses.activate
    def test_oauth_flow_error_from_auth(self, client):
        # mock response for SalesforceOAuth2.get_token()
        expected_response = {
            "access_token": "abc123",
            "id_token": "abc123",
            "token_type": "Bearer",
            "signature": "abc123",
            "issued_at": "12345",
            "scope": "web full refresh_token",
            "instance_url": "https://na15.salesforce.com",
            "id": "https://login.salesforce.com/id/abc/xyz",
            "refresh_token": "abc123",
        }
        responses.add(
            responses.POST,
            "https://login.salesforce.com/services/oauth2/token",
            status=http.client.OK,
            json=expected_response,
        )

        # call OAuth object on another thread - this spawns local httpd
        with httpd_thread(client):
            # simulate callback from browser
            with pytest.raises(urllib.error.HTTPError):
                urllib.request.urlopen(
                    client.client_config.redirect_uri
                    + "?error=123&error_description=broken"
                )

    @pytest.mark.skipif(
        sys.platform.startswith("win"), reason="setup differs from windows"
    )
    def test_create_httpd__port_already_in_use(self, client):
        with httpd_thread(client):
            with pytest.raises(
                OAuth2Error, match=PORT_IN_USE_ERR.format(DEFAULT_CONNECTED_APP_PORT)
            ):
                client._create_httpd()

    @mock.patch("cumulusci.oauth.client.HTTPServer")
    def test_create_httpd__other_OSError(self, HTTPServer, client):
        message = "generic error message"
        HTTPServer.side_effect = OSError(message)
        with pytest.raises(OSError, match=message):
            client._create_httpd()

    @responses.activate
    def test_oauth_flow_error_from_token(self, client):
        # mock response for OAuth2Client.get_access_token()
        responses.add(
            responses.POST,
            "https://login.salesforce.com/services/oauth2/token",
            status=http.client.FORBIDDEN,
        )

        # call OAuth object on another thread - this spawns local httpd
        with httpd_thread(client):
            # simulate callback from browser
            with pytest.raises(urllib.error.HTTPError):
                urllib.request.urlopen(client.client_config.redirect_uri + "?code=123")

    def test_validate_response__raises_error(self, client):
        response = Response()
        response.status_code = 400
        with pytest.raises(OAuth2Error):
            client.validate_response(response)


@pytest.fixture
def device_client_config():
    """Modified client_config for device auth flow"""
    client_config = {
        "auth_uri": "https://github.com/login/device/code",
        "client_id": "2a4bc3e5ce4f2c49a957",
        "token_uri": "https://github.com/login/oauth/access_token",
        "scope": "repo gist",
    }
    return OAuth2ClientConfig(**client_config)


@pytest.fixture
def device_user_code_resp():
    return {
        "device_code": "36482450e39b7f27d9a145a96898d29365a4e73f",
        "user_code": "3E15-9D06",
        "verification_uri": "https://github.com/login/device",
        "expires_in": 899,
        "interval": 5,
    }


@pytest.fixture
def device_device_config(device_user_code_resp):
    return OAuth2DeviceConfig(device_user_code_resp)


@pytest.fixture(autouse=True)
def configure_recording_mode():
    """Ignore --replace-vcr or integration tests fail."""
    pass


# Use run_code_without_recording for these or `make vcr` fails.
@pytest.mark.vcr()
def test_get_device_code(device_client_config, device_user_code_resp):
    response_dict: dict = get_device_code(device_client_config)
    expected = device_user_code_resp
    assert expected == response_dict


@pytest.mark.vcr()
def test_get_device_oauth_token(device_client_config, device_user_code_resp):
    """
    Note that the get_device_oauth_token includes a sleep while wating for the
    user to authorize the app. Mock `time.sleep` to regenerate the VCR cassette.
    """
    device_config = OAuth2DeviceConfig(**device_user_code_resp)
    response_dict = get_device_oauth_token(device_client_config, device_config)
    expected_response_dict = {
        "access_token": "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",
        "token_type": "bearer",
        "scope": "gist,repo",
    }
    assert expected_response_dict == response_dict
